from datetime import datetime
from random import randint
from string import ascii_letters

import json
import os
import re
import uuid
import time
import requests
import pytest
from types import SimpleNamespace
from typing import Type
from enum import Enum
from typing import Tuple

from labelbox import Dataset, DataRow
from labelbox import MediaType
from labelbox.orm import query
from labelbox.pagination import PaginatedCollection
from labelbox.schema.invite import Invite
from labelbox.schema.quality_mode import QualityMode
from labelbox.schema.queue_mode import QueueMode
from labelbox import Client

from labelbox import Dataset, DataRow
from labelbox import LabelingFrontend
from labelbox import OntologyBuilder, Tool, Option, Classification, MediaType
from labelbox.orm import query
from labelbox.pagination import PaginatedCollection
from labelbox.schema.annotation_import import LabelImport
from labelbox.schema.catalog import Catalog
from labelbox.schema.enums import AnnotationImportState
from labelbox.schema.invite import Invite
from labelbox.schema.quality_mode import QualityMode
from labelbox.schema.queue_mode import QueueMode
from labelbox.schema.user import User
from labelbox import Client

IMG_URL = "https://picsum.photos/200/300.jpg"
MASKABLE_IMG_URL = "https://storage.googleapis.com/labelbox-datasets/image_sample_data/2560px-Kitano_Street_Kobe01s5s4110.jpeg"
SMALL_DATASET_URL = "https://storage.googleapis.com/lb-artifacts-testing-public/sdk_integration_test/potato.jpeg"
DATA_ROW_PROCESSING_WAIT_TIMEOUT_SECONDS = 30
DATA_ROW_PROCESSING_WAIT_SLEEP_INTERNAL_SECONDS = 3
EPHEMERAL_BASE_URL = "http://lb-api-public"
IMAGE_URL = "https://storage.googleapis.com/diagnostics-demo-data/coco/COCO_train2014_000000000034.jpg"
EXTERNAL_ID = "my-image"

pytest_plugins = []


@pytest.fixture(scope="session")
def rand_gen():

    def gen(field_type):
        if field_type is str:
            return "".join(ascii_letters[randint(0,
                                                 len(ascii_letters) - 1)]
                           for _ in range(16))

        if field_type is datetime:
            return datetime.now()

        raise Exception("Can't random generate for field type '%r'" %
                        field_type)

    return gen


class Environ(Enum):
    LOCAL = 'local'
    PROD = 'prod'
    STAGING = 'staging'
    CUSTOM = 'custom'
    STAGING_EU = 'staging-eu'
    EPHEMERAL = 'ephemeral'  # Used for testing PRs with ephemeral environments


@pytest.fixture
def image_url() -> str:
    return MASKABLE_IMG_URL


@pytest.fixture
def external_id() -> str:
    return EXTERNAL_ID


def ephemeral_endpoint() -> str:
    return os.getenv('LABELBOX_TEST_BASE_URL', EPHEMERAL_BASE_URL)


def graphql_url(environ: str) -> str:
    if environ == Environ.LOCAL:
        return 'http://localhost:3000/api/graphql'
    elif environ == Environ.PROD:
        return 'https://api.labelbox.com/graphql'
    elif environ == Environ.STAGING:
        return 'https://api.lb-stage.xyz/graphql'
    elif environ == Environ.CUSTOM:
        graphql_api_endpoint = os.environ.get(
            'LABELBOX_TEST_GRAPHQL_API_ENDPOINT')
        if graphql_api_endpoint is None:
            raise Exception("Missing LABELBOX_TEST_GRAPHQL_API_ENDPOINT")
        return graphql_api_endpoint
    elif environ == Environ.EPHEMERAL:
        return f"{ephemeral_endpoint()}/graphql"
    return 'http://host.docker.internal:8080/graphql'


def rest_url(environ: str) -> str:
    if environ == Environ.LOCAL:
        return 'http://localhost:3000/api/v1'
    elif environ == Environ.PROD:
        return 'https://api.labelbox.com/api/v1'
    elif environ == Environ.STAGING:
        return 'https://api.lb-stage.xyz/api/v1'
    elif environ == Environ.CUSTOM:
        rest_api_endpoint = os.environ.get('LABELBOX_TEST_REST_API_ENDPOINT')
        if rest_api_endpoint is None:
            raise Exception("Missing LABELBOX_TEST_REST_API_ENDPOINT")
        return rest_api_endpoint
    elif environ == Environ.EPHEMERAL:
        return f"{ephemeral_endpoint()}/api/v1"
    return 'http://host.docker.internal:8080/api/v1'


def testing_api_key(environ: Environ) -> str:
    keys = [
        f"LABELBOX_TEST_API_KEY_{environ.value.upper()}",
        "LABELBOX_TEST_API_KEY",
        "LABELBOX_API_KEY"
    ]
    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            return value
    raise Exception("Cannot find API to use for tests")


def service_api_key() -> str:
    service_api_key = os.environ["SERVICE_API_KEY"]
    if service_api_key is None:
        raise Exception(
            "SERVICE_API_KEY is missing and needed for admin client")
    return service_api_key


class IntegrationClient(Client):

    def __init__(self, environ: str) -> None:
        api_url = graphql_url(environ)
        api_key = testing_api_key(environ)
        rest_endpoint = rest_url(environ)
        super().__init__(api_key,
                         api_url,
                         enable_experimental=True,
                         rest_endpoint=rest_endpoint)
        self.queries = []

    def execute(self, query=None, params=None, check_naming=True, **kwargs):
        if check_naming and query is not None:
            assert re.match(r"\s*(?:query|mutation) \w+PyApi",
                            query) is not None
        self.queries.append((query, params))
        if not kwargs.get('timeout'):
            kwargs['timeout'] = 30.0
        return super().execute(query, params, **kwargs)


class AdminClient(Client):

    def __init__(self, env):
        """
            The admin client creates organizations and users using admin api described here https://labelbox.atlassian.net/wiki/spaces/AP/pages/2206564433/Internal+Admin+APIs.
        """
        self._api_key = service_api_key()
        self._admin_endpoint = f"{ephemeral_endpoint()}/admin/v1"
        self._api_url = graphql_url(env)
        self._rest_endpoint = rest_url(env)

        super().__init__(self._api_key,
                         self._api_url,
                         enable_experimental=True,
                         rest_endpoint=self._rest_endpoint)

    def _create_organization(self) -> str:
        endpoint = f"{self._admin_endpoint}/organizations/"
        response = requests.post(
            endpoint,
            headers=self.headers,
            json={"name": f"Test Org {uuid.uuid4()}"},
        )

        data = response.json()
        if response.status_code not in [
                requests.codes.created, requests.codes.ok
        ]:
            raise Exception("Failed to create org, message: " +
                            str(data['message']))

        return data['id']

    def _create_user(self, organization_id=None) -> Tuple[str, str]:
        if organization_id is None:
            organization_id = self.organization_id

        endpoint = f"{self._admin_endpoint}/user-identities/"
        identity_id = f"e2e+{uuid.uuid4()}"

        response = requests.post(
            endpoint,
            headers=self.headers,
            json={
                "identityId": identity_id,
                "email": "email@email.com",
                "name": f"tester{uuid.uuid4()}",
                "verificationStatus": "VERIFIED",
            },
        )
        data = response.json()
        if response.status_code not in [
                requests.codes.created, requests.codes.ok
        ]:
            raise Exception("Failed to create user, message: " +
                            str(data['message']))

        user_identity_id = data['identityId']

        endpoint = f"{self._admin_endpoint}/organizations/{organization_id}/users/"
        response = requests.post(
            endpoint,
            headers=self.headers,
            json={
                "identityId": user_identity_id,
                "organizationRole": "Admin"
            },
        )

        data = response.json()
        if response.status_code not in [
                requests.codes.created, requests.codes.ok
        ]:
            raise Exception("Failed to create link user to org, message: " +
                            str(data['message']))

        user_id = data['id']

        endpoint = f"{self._admin_endpoint}/users/{user_id}/token"
        response = requests.get(
            endpoint,
            headers=self.headers,
        )
        data = response.json()
        if response.status_code not in [
                requests.codes.created, requests.codes.ok
        ]:
            raise Exception("Failed to create ephemeral user, message: " +
                            str(data['message']))

        token = data["token"]

        return user_id, token

    def create_api_key_for_user(self) -> str:
        organization_id = self._create_organization()
        _, user_token = self._create_user(organization_id)
        key_name = f"test-key+{uuid.uuid4()}"
        query = """mutation CreateApiKeyPyApi($name: String!) {
                createApiKey(data: {name: $name}) {
                    id
                    jwt
                }
            }
        """
        params = {"name": key_name}
        self.headers["Authorization"] = f"Bearer {user_token}"
        res = self.execute(query, params, error_log_key="errors")

        return res["createApiKey"]["jwt"]


class EphemeralClient(Client):

    def __init__(self, environ=Environ.EPHEMERAL):
        self.admin_client = AdminClient(environ)
        self.api_key = self.admin_client.create_api_key_for_user()
        api_url = graphql_url(environ)
        rest_endpoint = rest_url(environ)

        super().__init__(self.api_key,
                         api_url,
                         enable_experimental=True,
                         rest_endpoint=rest_endpoint)


@pytest.fixture
def ephmeral_client() -> EphemeralClient:
    return EphemeralClient


@pytest.fixture
def admin_client() -> AdminClient:
    return AdminClient


@pytest.fixture
def integration_client() -> IntegrationClient:
    return IntegrationClient


@pytest.fixture(scope="session")
def environ() -> Environ:
    """
    Checks environment variables for LABELBOX_ENVIRON to be
    'prod' or 'staging'
    Make sure to set LABELBOX_TEST_ENVIRON in .github/workflows/python-package.yaml
    """
    keys = [
        "LABELBOX_TEST_ENV",
        "LABELBOX_TEST_ENVIRON",
        "LABELBOX_ENV"
    ]
    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            return Environ(value)
    raise Exception(f'Missing env key in: {os.environ}')


def cancel_invite(client, invite_id):
    """
    Do not use. Only for testing.
    """
    query_str = """mutation CancelInvitePyApi($where: WhereUniqueIdInput!) {
            cancelInvite(where: $where) {id}}"""
    client.execute(query_str, {'where': {'id': invite_id}}, experimental=True)


def get_project_invites(client, project_id):
    """
    Do not use. Only for testing.
    """
    id_param = "projectId"
    query_str = """query GetProjectInvitationsPyApi($from: ID, $first: PageSize, $%s: ID!) {
        project(where: {id: $%s}) {id
        invites(from: $from, first: $first) { nodes { %s
        projectInvites { projectId projectRoleName } } nextCursor}}}
    """ % (id_param, id_param, query.results_query_part(Invite))
    return PaginatedCollection(client,
                               query_str, {id_param: project_id},
                               ['project', 'invites', 'nodes'],
                               Invite,
                               cursor_path=['project', 'invites', 'nextCursor'])


def get_invites(client):
    """
    Do not use. Only for testing.
    """
    query_str = """query GetOrgInvitationsPyApi($from: ID, $first: PageSize) {
            organization { id invites(from: $from, first: $first) {
                nodes { id createdAt organizationRoleName inviteeEmail } nextCursor }}}"""
    invites = PaginatedCollection(
        client,
        query_str, {}, ['organization', 'invites', 'nodes'],
        Invite,
        cursor_path=['organization', 'invites', 'nextCursor'],
        experimental=True)
    return invites


@pytest.fixture
def queries():
    return SimpleNamespace(cancel_invite=cancel_invite,
                           get_project_invites=get_project_invites,
                           get_invites=get_invites)


@pytest.fixture(scope="session")
def admin_client(environ: str):
    return AdminClient(environ)


@pytest.fixture(scope="session")
def client(environ: str):
    if environ == Environ.EPHEMERAL:
        return EphemeralClient()
    return IntegrationClient(environ)


@pytest.fixture(scope="session")
def pdf_url(client):
    pdf_url = client.upload_file('tests/assets/loremipsum.pdf')
    return {"row_data": {"pdf_url": pdf_url,}, "global_key": str(uuid.uuid4())}


@pytest.fixture(scope="session")
def pdf_entity_data_row(client):
    pdf_url = client.upload_file(
        'tests/assets/arxiv-pdf_data_99-word-token-pdfs_0801.3483.pdf')
    text_layer_url = client.upload_file(
        'tests/assets/arxiv-pdf_data_99-word-token-pdfs_0801.3483-lb-textlayer.json'
    )

    return {
        "row_data": {
            "pdf_url": pdf_url,
            "text_layer_url": text_layer_url
        },
        "global_key": str(uuid.uuid4())
    }


@pytest.fixture()
def conversation_entity_data_row(client, rand_gen):
    return {
        "row_data":
            "https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json",
        "global_key":
            f"https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json-{rand_gen(str)}",
    }


@pytest.fixture
def project(client, rand_gen):
    project = client.create_project(name=rand_gen(str),
                                    queue_mode=QueueMode.Batch,
                                    media_type=MediaType.Image)
    yield project
    project.delete()


@pytest.fixture
def consensus_project(client, rand_gen):
    project = client.create_project(name=rand_gen(str),
                                    quality_mode=QualityMode.Consensus,
                                    queue_mode=QueueMode.Batch,
                                    media_type=MediaType.Image)
    yield project
    project.delete()


@pytest.fixture
def model_config(client, rand_gen, valid_model_id):
    model_config = client.create_model_config(
        name=rand_gen(str),
        model_id=valid_model_id,
        inference_params={"param": "value"})
    yield model_config
    client.delete_model_config(model_config.uid)


@pytest.fixture
def consensus_project_with_batch(consensus_project, initial_dataset, rand_gen,
                                 image_url):
    project = consensus_project
    dataset = initial_dataset

    data_rows = []
    for _ in range(3):
        data_rows.append({
            DataRow.row_data: image_url,
            DataRow.global_key: str(uuid.uuid4())
        })
    task = dataset.create_data_rows(data_rows)
    task.wait_till_done()
    assert task.status == "COMPLETE"

    data_rows = list(dataset.data_rows())
    assert len(data_rows) == 3
    batch = project.create_batch(
        rand_gen(str),
        data_rows,  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )

    yield [project, batch, data_rows]
    batch.delete()


@pytest.fixture
def dataset(client, rand_gen):
    dataset = client.create_dataset(name=rand_gen(str))
    yield dataset
    dataset.delete()


@pytest.fixture(scope='function')
def unique_dataset(client, rand_gen):
    dataset = client.create_dataset(name=rand_gen(str))
    yield dataset
    dataset.delete()


@pytest.fixture
def small_dataset(dataset: Dataset):
    task = dataset.create_data_rows([
        {
            "row_data": SMALL_DATASET_URL,
            "external_id": "my-image"
        },
    ] * 2)
    task.wait_till_done()

    yield dataset


@pytest.fixture
def data_row(dataset, image_url, rand_gen):
    global_key = f"global-key-{rand_gen(str)}"
    task = dataset.create_data_rows([
        {
            "row_data": image_url,
            "external_id": "my-image",
            "global_key": global_key
        },
    ])
    task.wait_till_done()
    dr = dataset.data_rows().get_one()
    yield dr
    dr.delete()


@pytest.fixture
def data_row_and_global_key(dataset, image_url, rand_gen):
    global_key = f"global-key-{rand_gen(str)}"
    task = dataset.create_data_rows([
        {
            "row_data": image_url,
            "external_id": "my-image",
            "global_key": global_key
        },
    ])
    task.wait_till_done()
    dr = dataset.data_rows().get_one()
    yield dr, global_key
    dr.delete()


# can be used with
# @pytest.mark.parametrize('data_rows', [<count of data rows>], indirect=True)
# if omitted, count defaults to 1
@pytest.fixture
def data_rows(dataset, image_url, request, wait_for_data_row_processing,
              client):
    count = 1
    if hasattr(request, 'param'):
        count = request.param

    datarows = [
        dict(row_data=image_url, global_key=f"global-key-{uuid.uuid4()}")
        for _ in range(count)
    ]

    task = dataset.create_data_rows(datarows)
    task.wait_till_done()
    datarows = dataset.data_rows().get_many(count)
    for dr in dataset.data_rows():
        wait_for_data_row_processing(client, dr)

    yield datarows

    for datarow in datarows:
        datarow.delete()


@pytest.fixture
def iframe_url(environ) -> str:
    if environ in [Environ.PROD, Environ.LOCAL]:
        return 'https://editor.labelbox.com'
    elif environ == Environ.STAGING:
        return 'https://editor.lb-stage.xyz'


@pytest.fixture
def sample_image() -> str:
    path_to_video = 'tests/integration/media/sample_image.jpg'
    return path_to_video


@pytest.fixture
def sample_video() -> str:
    path_to_video = 'tests/integration/media/cat.mp4'
    return path_to_video


@pytest.fixture
def sample_bulk_conversation() -> list:
    path_to_conversation = 'tests/integration/media/bulk_conversation.json'
    with open(path_to_conversation) as json_file:
        conversations = json.load(json_file)
    return conversations


@pytest.fixture
def organization(client):
    # Must have at least one seat open in your org to run these tests
    org = client.get_organization()
    # Clean up before and after incase this wasn't run for some reason.
    for invite in get_invites(client):
        if "@labelbox.com" in invite.email:
            cancel_invite(client, invite.uid)
    yield org
    for invite in get_invites(client):
        if "@labelbox.com" in invite.email:
            cancel_invite(client, invite.uid)


@pytest.fixture
def configured_project_with_label(client, rand_gen, image_url, project, dataset,
                                  data_row, wait_for_label_processing):
    """Project with a connected dataset, having one datarow
    Project contains an ontology with 1 bbox tool
    Additionally includes a create_label method for any needed extra labels
    One label is already created and yielded when using fixture
    """
    project._wait_until_data_rows_are_processed(
        data_row_ids=[data_row.uid],
        wait_processing_max_seconds=DATA_ROW_PROCESSING_WAIT_TIMEOUT_SECONDS,
        sleep_interval=DATA_ROW_PROCESSING_WAIT_SLEEP_INTERNAL_SECONDS)

    project.create_batch(
        rand_gen(str),
        [data_row.uid],  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )
    ontology = _setup_ontology(project)
    label = _create_label(project, data_row, ontology,
                          wait_for_label_processing)
    yield [project, dataset, data_row, label]

    for label in project.labels():
        label.delete()


def _create_label(project, data_row, ontology, wait_for_label_processing):
    predictions = [{
        "uuid": str(uuid.uuid4()),
        "schemaId": ontology.tools[0].feature_schema_id,
        "dataRow": {
            "id": data_row.uid
        },
        "bbox": {
            "top": 20,
            "left": 20,
            "height": 50,
            "width": 50
        }
    }]

    def create_label():
        """ Ad-hoc function to create a LabelImport
        Creates a LabelImport task which will create a label
        """
        upload_task = LabelImport.create_from_objects(
            project.client, project.uid, f'label-import-{uuid.uuid4()}',
            predictions)
        upload_task.wait_until_done(sleep_time_seconds=5)
        assert upload_task.state == AnnotationImportState.FINISHED, "Label Import did not finish"
        assert len(
            upload_task.errors
        ) == 0, f"Label Import {upload_task.name} failed with errors {upload_task.errors}"

    project.create_label = create_label
    project.create_label()
    label = wait_for_label_processing(project)[0]
    return label


def _setup_ontology(project):
    editor = list(
        project.client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]
    ontology_builder = OntologyBuilder(tools=[
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class"),
    ])
    project.setup(editor, ontology_builder.asdict())
    # TODO: ontology may not be synchronous after setup. remove sleep when api is more consistent
    time.sleep(2)
    return OntologyBuilder.from_project(project)


@pytest.fixture
def big_dataset(dataset: Dataset):
    task = dataset.create_data_rows([
        {
            "row_data": IMAGE_URL,
            "external_id": EXTERNAL_ID
        },
    ] * 3)
    task.wait_till_done()

    yield dataset


@pytest.fixture
def configured_batch_project_with_label(project, dataset, data_row,
                                        wait_for_label_processing):
    """Project with a batch having one datarow
    Project contains an ontology with 1 bbox tool
    Additionally includes a create_label method for any needed extra labels
    One label is already created and yielded when using fixture
    """
    data_rows = [dr.uid for dr in list(dataset.data_rows())]
    project._wait_until_data_rows_are_processed(data_row_ids=data_rows,
                                                sleep_interval=3)
    project.create_batch("test-batch", data_rows)
    project.data_row_ids = data_rows

    ontology = _setup_ontology(project)
    label = _create_label(project, data_row, ontology,
                          wait_for_label_processing)

    yield [project, dataset, data_row, label]

    for label in project.labels():
        label.delete()


@pytest.fixture
def configured_batch_project_with_multiple_datarows(project, dataset, data_rows,
                                                    wait_for_label_processing):
    """Project with a batch having multiple datarows
    Project contains an ontology with 1 bbox tool
    Additionally includes a create_label method for any needed extra labels
    """
    global_keys = [dr.global_key for dr in data_rows]

    batch_name = f'batch {uuid.uuid4()}'
    project.create_batch(batch_name, global_keys=global_keys)

    ontology = _setup_ontology(project)
    for datarow in data_rows:
        _create_label(project, datarow, ontology, wait_for_label_processing)

    yield [project, dataset, data_rows]

    for label in project.labels():
        label.delete()


# NOTE this is nice heuristics, also there is this logic _wait_until_data_rows_are_processed in Project
#    in case we still have flakiness in the future, we can use it
@pytest.fixture
def wait_for_data_row_processing():
    """
    Do not use. Only for testing.

    Returns DataRow after waiting for it to finish processing media_attributes.
    Some tests, specifically ones that rely on label export, rely on
    DataRow be fully processed with media_attributes
    """

    def func(client, data_row, custom_check=None):
        """
        added check_updated_at because when a data_row is updated from say
        an image to pdf, it already has media_attributes and the loop does
        not wait for processing to a pdf
        """
        data_row_id = data_row.uid
        timeout_seconds = 60
        while True:
            data_row = client.get_data_row(data_row_id)
            passed_custom_check = not custom_check or custom_check(data_row)
            if data_row.media_attributes and passed_custom_check:
                return data_row
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for DataRow '{data_row_id}' to finish processing media_attributes"
                )
            time.sleep(2)

    return func


@pytest.fixture
def wait_for_label_processing():
    """
    Do not use. Only for testing.

    Returns project's labels as a list after waiting for them to finish processing.
    If `project.labels()` is called before label is fully processed,
    it may return an empty set
    """

    def func(project):
        timeout_seconds = 10
        while True:
            labels = list(project.labels())
            if len(labels) > 0:
                return labels
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for label for project '{project.uid}' to finish processing"
                )
            time.sleep(2)

    return func


@pytest.fixture
def initial_dataset(client, rand_gen):
    dataset = client.create_dataset(name=rand_gen(str))
    yield dataset

    dataset.delete()


@pytest.fixture
def video_data(client, rand_gen, video_data_row, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    data_row = dataset.create_data_row(video_data_row)
    data_row = wait_for_data_row_processing(client, data_row)
    data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


def create_video_data_row(rand_gen):
    return {
        "row_data":
            "https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4",
        "global_key":
            f"https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4-{rand_gen(str)}",
        "media_type":
            "VIDEO",
    }


@pytest.fixture
def video_data_100_rows(client, rand_gen, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    for _ in range(100):
        data_row = dataset.create_data_row(create_video_data_row(rand_gen))
        data_row = wait_for_data_row_processing(client, data_row)
        data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


@pytest.fixture()
def video_data_row(rand_gen):
    return create_video_data_row(rand_gen)


class ExportV2Helpers:

    @classmethod
    def run_project_export_v2_task(cls,
                                   project,
                                   num_retries=5,
                                   task_name=None,
                                   filters={},
                                   params={}):
        task = None
        params = params if params else {
            "project_details": True,
            "performance_details": False,
            "data_row_details": True,
            "label_details": True
        }
        while (num_retries > 0):
            task = project.export_v2(task_name=task_name,
                                     filters=filters,
                                     params=params)
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break
        return task.result

    @classmethod
    def run_dataset_export_v2_task(cls,
                                   dataset,
                                   num_retries=5,
                                   task_name=None,
                                   filters={},
                                   params={}):
        task = None
        params = params if params else {
            "performance_details": False,
            "label_details": True
        }
        while (num_retries > 0):
            task = dataset.export_v2(task_name=task_name,
                                     filters=filters,
                                     params=params)
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result

    @classmethod
    def run_catalog_export_v2_task(cls,
                                   client,
                                   num_retries=5,
                                   task_name=None,
                                   filters={},
                                   params={}):
        task = None
        params = params if params else {
            "performance_details": False,
            "label_details": True
        }
        catalog = client.get_catalog()
        while (num_retries > 0):

            task = catalog.export_v2(task_name=task_name,
                                     filters=filters,
                                     params=params)
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result


@pytest.fixture
def export_v2_test_helpers() -> Type[ExportV2Helpers]:
    return ExportV2Helpers()


@pytest.fixture
def big_dataset_data_row_ids(big_dataset: Dataset):
    export_task = big_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    yield [dr.json["data_row"]["id"] for dr in stream]


@pytest.fixture(scope='function')
def dataset_with_invalid_data_rows(unique_dataset: Dataset,
                                   upload_invalid_data_rows_for_dataset):
    upload_invalid_data_rows_for_dataset(unique_dataset)

    yield unique_dataset


@pytest.fixture
def upload_invalid_data_rows_for_dataset():

    def _upload_invalid_data_rows_for_dataset(dataset: Dataset):
        task = dataset.create_data_rows([
            {
                "row_data": 'gs://invalid-bucket/example.png',  # forbidden
                "external_id": "image-without-access.jpg"
            },
        ] * 2)
        task.wait_till_done()

    return _upload_invalid_data_rows_for_dataset


@pytest.fixture
def configured_project(project_with_empty_ontology, initial_dataset, rand_gen,
                       image_url):
    dataset = initial_dataset
    data_row_id = dataset.create_data_row(row_data=image_url).uid
    project = project_with_empty_ontology

    batch = project.create_batch(
        rand_gen(str),
        [data_row_id],  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = [data_row_id]

    yield project

    batch.delete()


@pytest.fixture
def project_with_empty_ontology(project):
    editor = list(
        project.client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]
    empty_ontology = {"tools": [], "classifications": []}
    project.setup(editor, empty_ontology)
    yield project


@pytest.fixture
def configured_project_with_complex_ontology(client, initial_dataset, rand_gen,
                                             image_url):
    project = client.create_project(name=rand_gen(str),
                                    queue_mode=QueueMode.Batch,
                                    media_type=MediaType.Image)
    dataset = initial_dataset
    data_row = dataset.create_data_row(row_data=image_url)
    data_row_ids = [data_row.uid]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids

    editor = list(
        project.client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]

    ontology = OntologyBuilder()
    tools = [
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class"),
        Tool(tool=Tool.Type.LINE, name="test-line-class"),
        Tool(tool=Tool.Type.POINT, name="test-point-class"),
        Tool(tool=Tool.Type.POLYGON, name="test-polygon-class"),
        Tool(tool=Tool.Type.NER, name="test-ner-class")
    ]

    options = [
        Option(value="first option answer"),
        Option(value="second option answer"),
        Option(value="third option answer")
    ]

    classifications = [
        Classification(class_type=Classification.Type.TEXT,
                       name="test-text-class"),
        Classification(class_type=Classification.Type.DROPDOWN,
                       name="test-dropdown-class",
                       options=options),
        Classification(class_type=Classification.Type.RADIO,
                       name="test-radio-class",
                       options=options),
        Classification(class_type=Classification.Type.CHECKLIST,
                       name="test-checklist-class",
                       options=options)
    ]

    for t in tools:
        for c in classifications:
            t.add_classification(c)
        ontology.add_tool(t)
    for c in classifications:
        ontology.add_classification(c)

    project.setup(editor, ontology.asdict())

    yield [project, data_row]
    project.delete()


@pytest.fixture
def embedding(client: Client):
    uuid_str = uuid.uuid4().hex
    embedding = client.create_embedding(f"sdk-int-{uuid_str}", 8)
    yield embedding
    embedding.delete()


@pytest.fixture
def valid_model_id():
    return "2c903542-d1da-48fd-9db1-8c62571bd3d2"



from unittest.mock import patch
import uuid
from labelbox import parser, Project
from labelbox.data.annotation_types.data.generic_data_row_data import GenericDataRowData
import pytest
import random
from labelbox.data.annotation_types.annotation import ObjectAnnotation
from labelbox.data.annotation_types.classification.classification import Checklist, ClassificationAnnotation, ClassificationAnswer, Radio
from labelbox.data.annotation_types.data.video import VideoData
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.rectangle import Rectangle, RectangleUnit
from labelbox.data.annotation_types.label import Label
from labelbox.data.annotation_types.data.text import TextData
from labelbox.data.annotation_types.ner import DocumentEntity, DocumentTextSelection
from labelbox.data.annotation_types.video import VideoObjectAnnotation

from labelbox.data.serialization import NDJsonConverter
from labelbox.exceptions import MALValidationError, UuidError
from labelbox.schema.bulk_import_request import BulkImportRequest
from labelbox.schema.enums import BulkImportRequestState
from labelbox.schema.annotation_import import LabelImport, MALPredictionImport
from labelbox.schema.media_type import MediaType
"""
- Here we only want to check that the uploads are calling the validation
- Then with unit tests we can check the types of errors raised
"""
#TODO: remove library once bulk import requests are removed

@pytest.mark.order(1)
def test_create_from_url(module_project):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"

    bulk_import_request = module_project.upload_annotations(name=name,
                                                            annotations=url,
                                                            validate=False)

    assert bulk_import_request.project() == module_project
    assert bulk_import_request.name == name
    assert bulk_import_request.input_file_url == url
    assert bulk_import_request.error_file_url is None
    assert bulk_import_request.status_file_url is None
    assert bulk_import_request.state == BulkImportRequestState.RUNNING


def test_validate_file(module_project):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    with pytest.raises(MALValidationError):
        module_project.upload_annotations(name=name,
                                          annotations=url,
                                          validate=True)
        #Schema ids shouldn't match


def test_create_from_objects(module_project: Project, predictions,
                             annotation_import_test_helpers):
    name = str(uuid.uuid4())

    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=predictions)

    assert bulk_import_request.project() == module_project
    assert bulk_import_request.name == name
    assert bulk_import_request.error_file_url is None
    assert bulk_import_request.status_file_url is None
    assert bulk_import_request.state == BulkImportRequestState.RUNNING
    annotation_import_test_helpers.assert_file_content(
        bulk_import_request.input_file_url, predictions)


def test_create_from_label_objects(module_project, predictions,
                                   annotation_import_test_helpers):
    name = str(uuid.uuid4())

    labels = list(NDJsonConverter.deserialize(predictions))
    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=labels)

    assert bulk_import_request.project() == module_project
    assert bulk_import_request.name == name
    assert bulk_import_request.error_file_url is None
    assert bulk_import_request.status_file_url is None
    assert bulk_import_request.state == BulkImportRequestState.RUNNING
    normalized_predictions = list(NDJsonConverter.serialize(labels))
    annotation_import_test_helpers.assert_file_content(
        bulk_import_request.input_file_url, normalized_predictions)


def test_create_from_local_file(tmp_path, predictions, module_project,
                                annotation_import_test_helpers):
    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(predictions, f)

    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=str(file_path), validate=False)

    assert bulk_import_request.project() == module_project
    assert bulk_import_request.name == name
    assert bulk_import_request.error_file_url is None
    assert bulk_import_request.status_file_url is None
    assert bulk_import_request.state == BulkImportRequestState.RUNNING
    annotation_import_test_helpers.assert_file_content(
        bulk_import_request.input_file_url, predictions)


def test_get(client, module_project):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    module_project.upload_annotations(name=name,
                               annotations=url,
                               validate=False)

    bulk_import_request = BulkImportRequest.from_name(
        client, project_id=module_project.uid, name=name)

    assert bulk_import_request.project() == module_project
    assert bulk_import_request.name == name
    assert bulk_import_request.input_file_url == url
    assert bulk_import_request.error_file_url is None
    assert bulk_import_request.status_file_url is None
    assert bulk_import_request.state == BulkImportRequestState.RUNNING


def test_validate_ndjson(tmp_path, module_project):
    file_name = f"broken.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        f.write("test")

    with pytest.raises(ValueError):
        module_project.upload_annotations(
            name="name", validate=True, annotations=str(file_path))


def test_validate_ndjson_uuid(tmp_path, module_project, predictions):
    file_name = f"repeat_uuid.ndjson"
    file_path = tmp_path / file_name
    repeat_uuid = predictions.copy()
    uid = str(uuid.uuid4())
    repeat_uuid[0]['uuid'] = uid
    repeat_uuid[1]['uuid'] = uid

    with file_path.open("w") as f:
        parser.dump(repeat_uuid, f)

    with pytest.raises(UuidError):
        module_project.upload_annotations(name="name",
                                              validate=True,
                                              annotations=str(file_path))

    with pytest.raises(UuidError):
        module_project.upload_annotations(name="name",
                                          validate=True,
                                          annotations=repeat_uuid)


@pytest.mark.skip("Slow test and uses a deprecated api endpoint for annotation imports")
def test_wait_till_done(rectangle_inference,
                        project):
    name = str(uuid.uuid4())
    url = project.client.upload_data(
        content=parser.dumps(rectangle_inference), sign=True)
    bulk_import_request = project.upload_annotations(
        name=name, annotations=url, validate=False)

    assert len(bulk_import_request.inputs) == 1
    bulk_import_request.wait_until_done()
    assert bulk_import_request.state == BulkImportRequestState.FINISHED

    # Check that the status files are being returned as expected
    assert len(bulk_import_request.errors) == 0
    assert len(bulk_import_request.inputs) == 1
    assert bulk_import_request.inputs[0]['uuid'] == rectangle_inference['uuid']
    assert len(bulk_import_request.statuses) == 1
    assert bulk_import_request.statuses[0]['status'] == 'SUCCESS'
    assert bulk_import_request.statuses[0]['uuid'] == rectangle_inference[
        'uuid']


def test_project_bulk_import_requests(module_project, predictions):
    result = module_project.bulk_import_requests()
    assert len(list(result)) == 0

    name = str(uuid.uuid4())
    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=predictions)
    bulk_import_request.wait_until_done()

    name = str(uuid.uuid4())
    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=predictions)
    bulk_import_request.wait_until_done()

    name = str(uuid.uuid4())
    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=predictions)
    bulk_import_request.wait_until_done()

    result = module_project.bulk_import_requests()
    assert len(list(result)) == 3


def test_delete(module_project, predictions):
    name = str(uuid.uuid4())
    
    bulk_import_requests = module_project.bulk_import_requests()
    [bulk_import_request.delete() for bulk_import_request in bulk_import_requests]
    
    bulk_import_request = module_project.upload_annotations(
        name=name, annotations=predictions)
    bulk_import_request.wait_until_done()
    all_import_requests = module_project.bulk_import_requests()
    assert len(list(all_import_requests)) == 1

    bulk_import_request.delete()
    all_import_requests = module_project.bulk_import_requests()
    assert len(list(all_import_requests)) == 0



from unittest.mock import MagicMock, patch
from labelbox.schema.export_task import (
    FileRetrieverByOffset,
    _TaskContext,
    _MetadataHeader,
    StreamType,
)


class TestFileRetrieverByOffset:

    def test_by_offset_from_start(self, generate_random_ndjson, mock_response):
        line_count = 10
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson) + "\n"

        mock_client = MagicMock()
        mock_client.execute = MagicMock(
            return_value={
                "task": {
                    "exportFileFromOffset": {
                        "offsets": {
                            "start": "0",
                            "end": len(file_content) - 1
                        },
                        "lines": {
                            "start": "0",
                            "end": str(line_count - 1)
                        },
                        "file": "http://some-url.com/file.ndjson",
                    }
                }
            })

        mock_ctx = _TaskContext(
            client=mock_client,
            task_id="task-id",
            stream_type=StreamType.RESULT,
            metadata_header=_MetadataHeader(total_size=len(file_content),
                                            total_lines=line_count),
        )

        with patch("requests.get", return_value=mock_response(file_content)):
            retriever = FileRetrieverByOffset(mock_ctx, 0)
            info, content = retriever.get_next_chunk()
            assert info.offsets.start == 0
            assert info.offsets.end == len(file_content) - 1
            assert info.lines.start == 0
            assert info.lines.end == line_count - 1
            assert info.file == "http://some-url.com/file.ndjson"
            assert content == file_content

    def test_by_offset_from_middle(self, generate_random_ndjson, mock_response):
        line_count = 10
        ndjson = generate_random_ndjson(line_count)
        file_content = "\n".join(ndjson) + "\n"

        mock_client = MagicMock()
        mock_client.execute = MagicMock(
            return_value={
                "task": {
                    "exportFileFromOffset": {
                        "offsets": {
                            "start": "0",
                            "end": len(file_content) - 1
                        },
                        "lines": {
                            "start": "0",
                            "end": str(line_count - 1)
                        },
                        "file": "http://some-url.com/file.ndjson",
                    }
                }
            })

        mock_ctx = _TaskContext(
            client=mock_client,
            task_id="task-id",
            stream_type=StreamType.RESULT,
            metadata_header=_MetadataHeader(total_size=len(file_content),
                                            total_lines=line_count),
        )

        line_start = 5
        skipped_bytes = 15
        current_offset = file_content.find(ndjson[line_start]) + skipped_bytes

        with patch("requests.get", return_value=mock_response(file_content)):
            retriever = FileRetrieverByOffset(mock_ctx, current_offset)
            info, content = retriever.get_next_chunk()
            assert info.offsets.start == current_offset
            assert info.offsets.end == len(file_content) - 1
            assert info.lines.start == 5
            assert info.lines.end == line_count - 1
            assert info.file == "http://some-url.com/file.ndjson"
            assert content == file_content[current_offset:]



from labelbox.typing_imports import Literal
from labelbox.utils import _NoCoercionMixin
from .base_data import BaseData


class AudioData(BaseData, _NoCoercionMixin):
    class_name: Literal["AudioData"] = "AudioData"


import json

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_nested():
    with open('tests/data/assets/ndjson/nested_import.json', 'r') as file:
        data = json.load(file)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data


def test_nested_name_only():
    with open('tests/data/assets/ndjson/nested_import_name_only.json',
              'r') as file:
        data = json.load(file)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data



from typing import Optional, Union, Type

from labelbox.data.annotation_types.data import ImageData, TextData
from labelbox.data.serialization.ndjson.base import DataRow, NDJsonBase
from labelbox.data.annotation_types.metrics.scalar import (
    ScalarMetric, ScalarMetricAggregation, ScalarMetricValue,
    ScalarMetricConfidenceValue)
from labelbox.data.annotation_types.metrics.confusion_matrix import (
    ConfusionMatrixAggregation, ConfusionMatrixMetric,
    ConfusionMatrixMetricValue, ConfusionMatrixMetricConfidenceValue)


class BaseNDMetric(NDJsonBase):
    metric_value: float
    feature_name: Optional[str] = None
    subclass_name: Optional[str] = None

    class Config:
        use_enum_values = True

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        for field in ['featureName', 'subclassName']:
            if res[field] is None:
                res.pop(field)
        return res


class NDConfusionMatrixMetric(BaseNDMetric):
    metric_value: Union[ConfusionMatrixMetricValue,
                        ConfusionMatrixMetricConfidenceValue]
    metric_name: str
    aggregation: ConfusionMatrixAggregation

    def to_common(self) -> ConfusionMatrixMetric:
        return ConfusionMatrixMetric(value=self.metric_value,
                                     metric_name=self.metric_name,
                                     feature_name=self.feature_name,
                                     subclass_name=self.subclass_name,
                                     aggregation=self.aggregation,
                                     extra={'uuid': self.uuid})

    @classmethod
    def from_common(
            cls, metric: ConfusionMatrixMetric,
            data: Union[TextData, ImageData]) -> "NDConfusionMatrixMetric":
        return cls(uuid=metric.extra.get('uuid'),
                   metric_value=metric.value,
                   metric_name=metric.metric_name,
                   feature_name=metric.feature_name,
                   subclass_name=metric.subclass_name,
                   aggregation=metric.aggregation,
                   data_row=DataRow(id=data.uid, global_key=data.global_key))


class NDScalarMetric(BaseNDMetric):
    metric_value: Union[ScalarMetricValue, ScalarMetricConfidenceValue]
    metric_name: Optional[str]
    aggregation: ScalarMetricAggregation = ScalarMetricAggregation.ARITHMETIC_MEAN

    def to_common(self) -> ScalarMetric:
        return ScalarMetric(value=self.metric_value,
                            metric_name=self.metric_name,
                            feature_name=self.feature_name,
                            subclass_name=self.subclass_name,
                            aggregation=self.aggregation,
                            extra={'uuid': self.uuid})

    @classmethod
    def from_common(cls, metric: ScalarMetric,
                    data: Union[TextData, ImageData]) -> "NDScalarMetric":
        return cls(uuid=metric.extra.get('uuid'),
                   metric_value=metric.value,
                   metric_name=metric.metric_name,
                   feature_name=metric.feature_name,
                   subclass_name=metric.subclass_name,
                   aggregation=metric.aggregation.value,
                   data_row=DataRow(id=data.uid, global_key=data.global_key))

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        # For backwards compatibility.
        if res['metricName'] is None:
            res.pop('metricName')
            res.pop('aggregation')
        return res


class NDMetricAnnotation:

    @classmethod
    def to_common(
        cls, annotation: Union[NDScalarMetric, NDConfusionMatrixMetric]
    ) -> Union[ScalarMetric, ConfusionMatrixMetric]:
        return annotation.to_common()

    @classmethod
    def from_common(
        cls, annotation: Union[ScalarMetric,
                               ConfusionMatrixMetric], data: Union[TextData,
                                                                   ImageData]
    ) -> Union[NDScalarMetric, NDConfusionMatrixMetric]:
        obj = cls.lookup_object(annotation)
        return obj.from_common(annotation, data)

    @staticmethod
    def lookup_object(
        annotation: Union[ScalarMetric, ConfusionMatrixMetric]
    ) -> Union[Type[NDScalarMetric], Type[NDConfusionMatrixMetric]]:
        result = {
            ScalarMetric: NDScalarMetric,
            ConfusionMatrixMetric: NDConfusionMatrixMetric,
        }.get(type(annotation))
        if result is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation)}`")
        return result



import numpy as np
import pytest

from labelbox.data.annotation_types import VideoData
from labelbox import pydantic_compat


def test_validate_schema():
    with pytest.raises(pydantic_compat.ValidationError):
        data = VideoData()


def test_frames():
    data = {
        x: (np.random.random((32, 32, 3)) * 255).astype(np.uint8)
        for x in range(5)
    }
    video_data = VideoData(frames=data)
    for idx, frame in video_data.frame_generator():
        assert idx in data
        assert np.all(frame == data[idx])


def test_file_path():
    path = 'tests/integration/media/cat.mp4'
    raster_data = VideoData(file_path=path)

    with pytest.raises(ValueError):
        raster_data[0]

    raster_data.load_frames()
    raster_data[0]

    frame_indices = list(raster_data.frames.keys())
    # 29 frames
    assert set(frame_indices) == set(list(range(28)))


def test_file_url():
    url = "http://commondatastorage.googleapis.com/gtv-videos-bucket/sample/ForBiggerMeltdowns.mp4"
    raster_data = VideoData(url=url)

    with pytest.raises(ValueError):
        raster_data[0]

    raster_data.load_frames()
    raster_data[0]

    frame_indices = list(raster_data.frames.keys())
    # 362 frames
    assert set(frame_indices) == set(list(range(361)))


def test_ref():
    external_id = "external_id"
    uid = "uid"
    data = {
        x: (np.random.random((32, 32, 3)) * 255).astype(np.uint8)
        for x in range(5)
    }
    metadata = []
    media_attributes = {}
    data = VideoData(frames=data,
                     external_id=external_id,
                     uid=uid,
                     metadata=metadata,
                     media_attributes=media_attributes)
    assert data.external_id == external_id
    assert data.uid == uid
    assert data.media_attributes == media_attributes
    assert data.metadata == metadata



from abc import ABC
from typing import Dict, Optional, Any, Union

from labelbox import pydantic_compat

ConfidenceValue = pydantic_compat.confloat(ge=0, le=1)

MIN_CONFIDENCE_SCORES = 2
MAX_CONFIDENCE_SCORES = 15


class BaseMetric(pydantic_compat.BaseModel, ABC):
    value: Union[Any, Dict[float, Any]]
    feature_name: Optional[str] = None
    subclass_name: Optional[str] = None
    extra: Dict[str, Any] = {}

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        return {k: v for k, v in res.items() if v is not None}

    @pydantic_compat.validator('value')
    def validate_value(cls, value):
        if isinstance(value, Dict):
            if not (MIN_CONFIDENCE_SCORES <= len(value) <=
                    MAX_CONFIDENCE_SCORES):
                raise pydantic_compat.ValidationError([
                    pydantic_compat.ErrorWrapper(ValueError(
                        "Number of confidence scores must be greater"
                        f" than or equal to {MIN_CONFIDENCE_SCORES} and"
                        f" less than or equal to {MAX_CONFIDENCE_SCORES}. Found {len(value)}"
                    ),
                                                 loc='value')
                ], cls)
        return value



import json
import pytest
import collections.abc
from labelbox import DataRow
from labelbox.schema.data_row_metadata import DataRowMetadataField

INTEGRATION_SNAPSHOT_DIRECTORY = "tests/integration/snapshots"
TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"


def test_task_errors(dataset, image_url, snapshot):
    client = dataset.client
    task = dataset.create_data_rows([
        {
            DataRow.row_data:
                image_url,
            DataRow.metadata_fields: [
                DataRowMetadataField(schema_id=TEXT_SCHEMA_ID,
                                     value='some msg'),
                DataRowMetadataField(schema_id=TEXT_SCHEMA_ID,
                                     value='some msg 2')
            ]
        },
    ])

    assert task in client.get_user().created_tasks()
    task.wait_till_done()

    assert len(task.failed_data_rows) == 1
    assert "A schemaId can only be specified once per DataRow : [cko8s9r5v0001h2dk9elqdidh]" in task.failed_data_rows[
        0]['message']
    assert len(task.failed_data_rows[0]['failedDataRows'][0]['metadata']) == 2


def test_task_success_json(dataset, image_url, snapshot):
    client = dataset.client
    task = dataset.create_data_rows([
        {
            DataRow.row_data: image_url,
        },
    ])
    assert task in client.get_user().created_tasks()
    task.wait_till_done()
    assert task.status == "COMPLETE"
    assert task.errors is None
    assert task.result is not None
    assert isinstance(task.result, collections.abc.Sequence)
    assert task.result_url is not None
    assert isinstance(task.result_url, str)
    task_result = task.result[0]
    assert 'id' in task_result and isinstance(task_result['id'], str)
    assert 'row_data' in task_result and isinstance(task_result['row_data'],
                                                    str)
    snapshot.snapshot_dir = INTEGRATION_SNAPSHOT_DIRECTORY
    task_result['id'] = 'DUMMY_ID'
    task_result['row_data'] = 'https://dummy.url'
    snapshot.assert_match(json.dumps(task_result),
                          'test_task.test_task_success_json.json')
    assert len(task.result)



from labelbox import LabelingFrontend


def test_get_labeling_frontends(client):
    filtered_frontends = list(
        client.get_labeling_frontends(where=LabelingFrontend.name == 'Editor'))
    assert len(filtered_frontends)


def test_labeling_frontend_connecting_to_project(project):
    client = project.client
    default_labeling_frontend = next(
        client.get_labeling_frontends(where=LabelingFrontend.name == "Editor"))

    assert project.labeling_frontend(
    ) == default_labeling_frontend  # we now have a default labeling frontend

    frontend = list(project.client.get_labeling_frontends())[0]
    project.labeling_frontend.connect(frontend)

    project.labeling_frontend.connect(frontend)
    assert project.labeling_frontend() == default_labeling_frontend

    project.labeling_frontend.disconnect(frontend)
    assert project.labeling_frontend() == None



from typing import Optional

from labelbox import pydantic_compat

from .types import Cuid


class FeatureSchema(pydantic_compat.BaseModel):
    """
    Class that represents a feature schema.
    Could be a annotation, a subclass, or an option.
    Schema ids might not be known when constructing these objects so both a name and schema id are valid.
    """
    name: Optional[str] = None
    feature_schema_id: Optional[Cuid] = None

    @pydantic_compat.root_validator
    def must_set_one(cls, values):
        if values['feature_schema_id'] is None and values['name'] is None:
            raise ValueError(
                "Must set either feature_schema_id or name for all feature schemas"
            )
        return values

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if 'name' in res and res['name'] is None:
            res.pop('name')
        if 'featureSchemaId' in res and res['featureSchemaId'] is None:
            res.pop('featureSchemaId')
        return res



from typing import Optional

from labelbox import pydantic_compat

from labelbox.utils import camel_case
from ...annotation_types.types import Cuid


class LBV1Feature(pydantic_compat.BaseModel):
    keyframe: Optional[bool] = None
    title: str = None
    value: Optional[str] = None
    schema_id: Optional[Cuid] = None
    feature_id: Optional[Cuid] = None

    @pydantic_compat.root_validator
    def check_ids(cls, values):
        if values.get('value') is None:
            values['value'] = values['title']
        return values

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        # This means these are no video frames ..
        if self.keyframe is None:
            res.pop('keyframe')
        return res

    class Config:
        allow_population_by_field_name = True
        alias_generator = camel_case



import itertools
import uuid
from labelbox.schema.annotation_import import AnnotationImport, MALPredictionImport
from labelbox.schema.media_type import MediaType
import pytest
from unittest.mock import patch


@patch('labelbox.schema.annotation_import.ANNOTATION_PER_LABEL_LIMIT', 1)
def test_above_annotation_limit_on_single_import_on_single_data_row(annotations_by_media_type):

    annotations_ndjson =  list(itertools.chain.from_iterable(annotations_by_media_type[MediaType.Image]))
    data_row_id = annotations_ndjson[0]["dataRow"]["id"]

    data_row_annotations = [annotation for annotation in annotations_ndjson if annotation["dataRow"]["id"] == data_row_id and "bbox" in annotation]
    
    with pytest.raises(ValueError):
        AnnotationImport._validate_data_rows([data_row_annotations[0]]*2)


@patch('labelbox.schema.annotation_import.ANNOTATION_PER_LABEL_LIMIT', 1)
def test_above_annotation_limit_divided_among_different_rows(annotations_by_media_type):

    annotations_ndjson =  list(itertools.chain.from_iterable(annotations_by_media_type[MediaType.Image]))
    data_row_id = annotations_ndjson[0]["dataRow"]["id"]
    
    first_data_row_annotation = [annotation for annotation in annotations_ndjson if annotation["dataRow"]["id"] == data_row_id and "bbox" in annotation][0]
    
    second_data_row_annotation = first_data_row_annotation.copy()
    second_data_row_annotation["dataRow"]["id"] == "data_row_id_2"
    
    with pytest.raises(ValueError):
        AnnotationImport._validate_data_rows([first_data_row_annotation, second_data_row_annotation]*2)


from labelbox.utils import _CamelCaseMixin

from labelbox import pydantic_compat

from datetime import datetime
from typing import Dict


class Model(_CamelCaseMixin, pydantic_compat.BaseModel):
    id: str
    description: str
    inference_params_json_schema: Dict
    name: str
    ontology_id: str
    created_at: datetime


MODEL_FIELD_NAMES = list(Model.schema()['properties'].keys())



import json

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter
from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_metric():
    with open('tests/data/assets/ndjson/metric_import.json', 'r') as file:
        data = json.load(file)

    label_list = list(NDJsonConverter.deserialize(data))
    reserialized = list(NDJsonConverter.serialize(label_list))
    assert reserialized == data

    # Just make sure that this doesn't break
    list(LBV1Converter.serialize(label_list))


def test_custom_scalar_metric():
    with open('tests/data/assets/ndjson/custom_scalar_import.json',
              'r') as file:
        data = json.load(file)

    label_list = list(NDJsonConverter.deserialize(data))
    reserialized = list(NDJsonConverter.serialize(label_list))
    assert json.dumps(reserialized,
                      sort_keys=True) == json.dumps(data, sort_keys=True)

    # Just make sure that this doesn't break
    list(LBV1Converter.serialize(label_list))


def test_custom_confusion_matrix_metric():
    with open('tests/data/assets/ndjson/custom_confusion_matrix_import.json',
              'r') as file:
        data = json.load(file)

    label_list = list(NDJsonConverter.deserialize(data))
    reserialized = list(NDJsonConverter.serialize(label_list))
    assert json.dumps(reserialized,
                      sort_keys=True) == json.dumps(data, sort_keys=True)

    # Just make sure that this doesn't break
    list(LBV1Converter.serialize(label_list))



from .audio import AudioData
from .conversation import ConversationData
from .dicom import DicomData
from .document import DocumentData
from .html import HTMLData
from .raster import ImageData
from .raster import MaskData
from .text import TextData
from .video import VideoData
from .llm_prompt_response_creation import LlmPromptResponseCreationData
from .llm_prompt_creation import LlmPromptCreationData
from .llm_response_creation import LlmResponseCreationData


from pytest_cases import fixture_ref
from pytest_cases import parametrize, fixture_ref

from labelbox.data.metrics.confusion_matrix.confusion_matrix import feature_confusion_matrix_metric


@parametrize("tool_examples", [
    fixture_ref('polygon_pairs'),
    fixture_ref('rectangle_pairs'),
    fixture_ref('mask_pairs'),
    fixture_ref('line_pairs'),
    fixture_ref('point_pairs'),
    fixture_ref('ner_pairs')
])
def test_overlapping_objects(tool_examples):
    for example in tool_examples:
        for include_subclasses, expected_attr_name in [[
                True, 'expected'
        ], [False, 'expected_without_subclasses']]:
            metrics = feature_confusion_matrix_metric(
                example.ground_truths,
                example.predictions,
                include_subclasses=include_subclasses)

            metrics = {r.feature_name: list(r.value) for r in metrics}
            if len(getattr(example, expected_attr_name)) == 0:
                assert len(metrics) == 0
            else:
                assert metrics == getattr(
                    example, expected_attr_name
                ), f"{example.predictions},{example.ground_truths}"


@parametrize("tool_examples",
             [fixture_ref('checklist_pairs'),
              fixture_ref('radio_pairs')])
def test_overlapping_classifications(tool_examples):
    for example in tool_examples:

        metrics = feature_confusion_matrix_metric(example.ground_truths,
                                                  example.predictions)

        metrics = {r.feature_name: list(r.value) for r in metrics}
        if len(example.expected) == 0:
            assert len(metrics) == 0
        else:
            assert metrics == example.expected, f"{example.predictions},{example.ground_truths}"



from typing import Any, Dict, List, Union, Optional

from labelbox import pydantic_compat
from labelbox.data.mixins import ConfidenceMixin, CustomMetric, CustomMetricsMixin
from labelbox.data.serialization.ndjson.base import DataRow, NDAnnotation

from labelbox.utils import camel_case
from ...annotation_types.annotation import ClassificationAnnotation
from ...annotation_types.video import VideoClassificationAnnotation
from ...annotation_types.llm_prompt_response.prompt import PromptClassificationAnnotation, PromptText
from ...annotation_types.classification.classification import ClassificationAnswer, Dropdown, Text, Checklist, Radio
from ...annotation_types.types import Cuid
from ...annotation_types.data import TextData, VideoData, ImageData


class NDAnswer(ConfidenceMixin, CustomMetricsMixin):
    name: Optional[str] = None
    schema_id: Optional[Cuid] = None
    classifications: Optional[List['NDSubclassificationType']] = []

    @pydantic_compat.root_validator()
    def must_set_one(cls, values):
        if ('schema_id' not in values or values['schema_id']
                is None) and ('name' not in values or values['name'] is None):
            raise ValueError("Schema id or name are not set. Set either one.")
        return values

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if 'name' in res and res['name'] is None:
            res.pop('name')
        if 'schemaId' in res and res['schemaId'] is None:
            res.pop('schemaId')
        if self.classifications is None or len(self.classifications) == 0:
            res.pop('classifications')
        else:
            res['classifications'] = [
                c.dict(*args, **kwargs) for c in self.classifications
            ]
        return res

    class Config:
        allow_population_by_field_name = True
        alias_generator = camel_case


class FrameLocation(pydantic_compat.BaseModel):
    end: int
    start: int


class VideoSupported(pydantic_compat.BaseModel):
    # Note that frames are only allowed as top level inferences for video
    frames: Optional[List[FrameLocation]] = None

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        # This means these are no video frames ..
        if self.frames is None:
            res.pop('frames')
        return res


class NDTextSubclass(NDAnswer):
    answer: str

    def to_common(self) -> Text:
        return Text(answer=self.answer,
                    confidence=self.confidence,
                    custom_metrics=self.custom_metrics)

    @classmethod
    def from_common(cls, text: Text, name: str,
                    feature_schema_id: Cuid) -> "NDTextSubclass":
        return cls(
            answer=text.answer,
            name=name,
            schema_id=feature_schema_id,
            confidence=text.confidence,
            custom_metrics=text.custom_metrics,
        )


class NDChecklistSubclass(NDAnswer):
    answer: List[NDAnswer] = pydantic_compat.Field(..., alias='answers')

    def to_common(self) -> Checklist:

        return Checklist(answer=[
            ClassificationAnswer(name=answer.name,
                                 feature_schema_id=answer.schema_id,
                                 confidence=answer.confidence,
                                 classifications=[
                                     NDSubclassification.to_common(annot)
                                     for annot in answer.classifications
                                 ],
                                 custom_metrics=answer.custom_metrics)
            for answer in self.answer
        ])

    @classmethod
    def from_common(cls, checklist: Checklist, name: str,
                    feature_schema_id: Cuid) -> "NDChecklistSubclass":
        return cls(answer=[
            NDAnswer(name=answer.name,
                     schema_id=answer.feature_schema_id,
                     confidence=answer.confidence,
                     classifications=[
                         NDSubclassification.from_common(annot)
                         for annot in answer.classifications
                     ],
                     custom_metrics=answer.custom_metrics)
            for answer in checklist.answer
        ],
                   name=name,
                   schema_id=feature_schema_id)

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if 'answers' in res:
            res['answer'] = res.pop('answers')
        return res


class NDRadioSubclass(NDAnswer):
    answer: NDAnswer

    def to_common(self) -> Radio:
        return Radio(answer=ClassificationAnswer(
            name=self.answer.name,
            feature_schema_id=self.answer.schema_id,
            confidence=self.answer.confidence,
            classifications=[
                NDSubclassification.to_common(annot)
                for annot in self.answer.classifications
            ],
            custom_metrics=self.answer.custom_metrics))

    @classmethod
    def from_common(cls, radio: Radio, name: str,
                    feature_schema_id: Cuid) -> "NDRadioSubclass":
        return cls(answer=NDAnswer(name=radio.answer.name,
                                   schema_id=radio.answer.feature_schema_id,
                                   confidence=radio.answer.confidence,
                                   classifications=[
                                       NDSubclassification.from_common(annot)
                                       for annot in radio.answer.classifications
                                   ],
                                   custom_metrics=radio.answer.custom_metrics),
                   name=name,
                   schema_id=feature_schema_id)


class NDPromptTextSubclass(NDAnswer):
    answer: str

    def to_common(self) -> PromptText:
        return PromptText(answer=self.answer,
                    confidence=self.confidence,
                    custom_metrics=self.custom_metrics)

    @classmethod
    def from_common(cls, prompt_text: PromptText, name: str,
                    feature_schema_id: Cuid) -> "NDPromptTextSubclass":
        return cls(
            answer=prompt_text.answer,
            name=name,
            schema_id=feature_schema_id,
            confidence=prompt_text.confidence,
            custom_metrics=prompt_text.custom_metrics,
        )


# ====== End of subclasses


class NDText(NDAnnotation, NDTextSubclass):

    @classmethod
    def from_common(cls,
                    uuid: str,
                    text: Text,
                    name: str,
                    feature_schema_id: Cuid,
                    extra: Dict[str, Any],
                    data: Union[TextData, ImageData],
                    message_id: str,
                    confidence: Optional[float] = None) -> "NDText":
        return cls(
            answer=text.answer,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            message_id=message_id,
            confidence=text.confidence,
            custom_metrics=text.custom_metrics,
        )


class NDChecklist(NDAnnotation, NDChecklistSubclass, VideoSupported):

    @classmethod
    def from_common(
            cls,
            uuid: str,
            checklist: Checklist,
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[VideoData, TextData, ImageData],
            message_id: str,
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None
    ) -> "NDChecklist":

        return cls(answer=[
            NDAnswer(name=answer.name,
                     schema_id=answer.feature_schema_id,
                     confidence=answer.confidence,
                     classifications=[
                         NDSubclassification.from_common(annot)
                         for annot in answer.classifications
                     ],
                     custom_metrics=answer.custom_metrics)
            for answer in checklist.answer
        ],
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   frames=extra.get('frames'),
                   message_id=message_id,
                   confidence=confidence)


class NDRadio(NDAnnotation, NDRadioSubclass, VideoSupported):

    @classmethod
    def from_common(
        cls,
        uuid: str,
        radio: Radio,
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: Union[VideoData, TextData, ImageData],
        message_id: str,
        confidence: Optional[float] = None,
    ) -> "NDRadio":
        return cls(answer=NDAnswer(name=radio.answer.name,
                                   schema_id=radio.answer.feature_schema_id,
                                   confidence=radio.answer.confidence,
                                   classifications=[
                                       NDSubclassification.from_common(annot)
                                       for annot in radio.answer.classifications
                                   ],
                                   custom_metrics=radio.answer.custom_metrics),
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   frames=extra.get('frames'),
                   message_id=message_id,
                   confidence=confidence)
        
        
class NDPromptText(NDAnnotation, NDPromptTextSubclass):
    
    @classmethod
    def from_common(
        cls,
        uuid: str,
        text: PromptText,
        name,
        data: Dict,
        feature_schema_id: Cuid,
        confidence: Optional[float] = None
    ) -> "NDPromptText":
        return cls(
            answer=text.answer,
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            name=name,
            schema_id=feature_schema_id,
            uuid=uuid,
            confidence=text.confidence,
            custom_metrics=text.custom_metrics)


class NDSubclassification:

    @classmethod
    def from_common(
        cls, annotation: ClassificationAnnotation
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        classify_obj = cls.lookup_subclassification(annotation)
        if classify_obj is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation.value)}`"
            )
        return classify_obj.from_common(annotation.value, annotation.name,
                                        annotation.feature_schema_id)

    @staticmethod
    def to_common(
            annotation: "NDClassificationType") -> ClassificationAnnotation:
        return ClassificationAnnotation(value=annotation.to_common(),
                                        name=annotation.name,
                                        feature_schema_id=annotation.schema_id)

    @staticmethod
    def lookup_subclassification(
        annotation: ClassificationAnnotation
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        if isinstance(annotation.value, Dropdown):
            raise TypeError("Dropdowns are not supported for MAL.")
        return {
            Text: NDTextSubclass,
            Checklist: NDChecklistSubclass,
            Radio: NDRadioSubclass,
        }.get(type(annotation.value))


class NDClassification:

    @staticmethod
    def to_common(
        annotation: "NDClassificationType"
    ) -> Union[ClassificationAnnotation, VideoClassificationAnnotation]:
        common = ClassificationAnnotation(
            value=annotation.to_common(),
            name=annotation.name,
            feature_schema_id=annotation.schema_id,
            extra={'uuid': annotation.uuid},
            message_id=annotation.message_id,
            confidence=annotation.confidence,
        )

        if getattr(annotation, 'frames', None) is None:
            return [common]
        results = []
        for frame in annotation.frames:
            for idx in range(frame.start, frame.end + 1, 1):
                results.append(
                    VideoClassificationAnnotation(frame=idx, **common.dict()))
        return results

    @classmethod
    def from_common(
        cls, annotation: Union[ClassificationAnnotation,
                               VideoClassificationAnnotation],
        data: Union[VideoData, TextData, ImageData]
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        classify_obj = cls.lookup_classification(annotation)
        if classify_obj is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation.value)}`"
            )
        return classify_obj.from_common(str(annotation._uuid), annotation.value,
                                        annotation.name,
                                        annotation.feature_schema_id,
                                        annotation.extra, data,
                                        annotation.message_id,
                                        annotation.confidence)

    @staticmethod
    def lookup_classification(
        annotation: Union[ClassificationAnnotation,
                          VideoClassificationAnnotation]
    ) -> Union[NDText, NDChecklist, NDRadio]:
        if isinstance(annotation.value, Dropdown):
            raise TypeError("Dropdowns are not supported for MAL.")
        return {
            Text: NDText,
            Checklist: NDChecklist,
            Radio: NDRadio
        }.get(type(annotation.value))

class NDPromptClassification:

    @staticmethod
    def to_common(
        annotation: "NDPromptClassificationType"
    ) -> Union[PromptClassificationAnnotation]:
        common = PromptClassificationAnnotation(
            value=annotation,
            name=annotation.name,
            feature_schema_id=annotation.schema_id,
            extra={'uuid': annotation.uuid},
            confidence=annotation.confidence,
        )

        return common

    @classmethod
    def from_common(
        cls, annotation: Union[PromptClassificationAnnotation],
        data: Union[VideoData, TextData, ImageData]
    ) -> Union[NDTextSubclass, NDChecklistSubclass, NDRadioSubclass]:
        return NDPromptText.from_common(str(annotation._uuid), annotation.value,
                                        annotation.name,
                                        data,
                                        annotation.feature_schema_id,
                                        annotation.confidence)


# Make sure to keep NDChecklistSubclass prior to NDRadioSubclass in the list,
# otherwise list of answers gets parsed by NDRadio whereas NDChecklist must be used
NDSubclassificationType = Union[NDChecklistSubclass, NDRadioSubclass,
                                NDTextSubclass]

NDAnswer.update_forward_refs()
NDChecklistSubclass.update_forward_refs()
NDChecklist.update_forward_refs()
NDRadioSubclass.update_forward_refs()
NDRadio.update_forward_refs()
NDText.update_forward_refs()
NDPromptText.update_forward_refs()
NDTextSubclass.update_forward_refs()

# Make sure to keep NDChecklist prior to NDRadio in the list,
# otherwise list of answers gets parsed by NDRadio whereas NDChecklist must be used
NDClassificationType = Union[NDChecklist, NDRadio, NDText]
NDPromptClassificationType = Union[NDPromptText]


from enum import Enum


class QueueMode(str, Enum):
    Batch = "BATCH"
    Dataset = "DATA_SET"

    @classmethod
    def _missing_(cls, value):
        # Parses the deprecated "CATALOG" value back to QueueMode.Batch.
        if value == "CATALOG":
            return QueueMode.Batch



from types import SimpleNamespace
from uuid import uuid4

import numpy as np
import pytest

from labelbox.data.annotation_types import (LabelGenerator, ObjectAnnotation,
                                            ImageData, MaskData, Line, Mask,
                                            Point, Label)
from labelbox import OntologyBuilder, Tool


@pytest.fixture
def list_of_labels():
    return [Label(data=ImageData(url="http://someurl")) for _ in range(5)]


@pytest.fixture
def signer():

    def get_signer(uuid):
        return lambda x: uuid

    return get_signer


class FakeDataset:

    def __init__(self):
        self.uid = "ckrb4tgm51xl10ybc7lv9ghm7"
        self.exports = []

    def create_data_row(self, row_data, external_id=None):
        if external_id is None:
            external_id = "an external_id"
        return SimpleNamespace(uid=self.uid, external_id=external_id)

    def create_data_rows(self, args):
        for arg in args:
            self.exports.append(
                SimpleNamespace(row_data=arg['row_data'],
                                external_id=arg['external_id'],
                                uid=self.uid))
        return self

    def wait_till_done(self):
        pass

    def export_data_rows(self):
        for export in self.exports:
            yield export


def test_generator(list_of_labels):
    generator = LabelGenerator([list_of_labels[0]])

    assert next(generator) == list_of_labels[0]
    with pytest.raises(StopIteration):
        next(generator)


def test_conversion(list_of_labels):
    generator = LabelGenerator(list_of_labels)
    label_collection = list(generator)
    assert len(label_collection) == len(list_of_labels)
    assert [x for x in label_collection] == list_of_labels


def test_adding_schema_ids():
    name = "line_feature"
    label = Label(
        data=ImageData(arr=np.ones((32, 32, 3), dtype=np.uint8)),
        annotations=[
            ObjectAnnotation(
                value=Line(
                    points=[Point(x=1, y=2), Point(x=2, y=2)]),
                name=name,
            )
        ])
    feature_schema_id = "expected_id"
    ontology = OntologyBuilder(tools=[
        Tool(Tool.Type.LINE, name=name, feature_schema_id=feature_schema_id)
    ])
    generator = LabelGenerator([label]).assign_feature_schema_ids(ontology)
    assert next(generator).annotations[0].feature_schema_id == feature_schema_id


def test_adding_urls(signer):
    label = Label(data=ImageData(arr=np.random.random((32, 32,
                                                       3)).astype(np.uint8)),
                  annotations=[])
    uuid = str(uuid4())
    generator = LabelGenerator([label]).add_url_to_data(signer(uuid))
    assert label.data.url != uuid
    assert next(generator).data.url == uuid
    assert label.data.url == uuid


def test_adding_to_dataset(signer):
    dataset = FakeDataset()
    label = Label(data=ImageData(arr=np.random.random((32, 32,
                                                       3)).astype(np.uint8)),
                  annotations=[])
    uuid = str(uuid4())
    generator = LabelGenerator([label]).add_to_dataset(dataset, signer(uuid))
    assert label.data.url != uuid
    generated_label = next(generator)
    assert generated_label.data.url == uuid
    assert generated_label.data.external_id != None
    assert generated_label.data.uid == dataset.uid
    assert label.data.url == uuid


def test_adding_to_masks(signer):
    label = Label(
        data=ImageData(arr=np.random.random((32, 32, 3)).astype(np.uint8)),
        annotations=[
            ObjectAnnotation(name="1234",
                             value=Mask(mask=MaskData(
                                 arr=np.random.random((32, 32,
                                                       3)).astype(np.uint8)),
                                        color=[255, 255, 255]))
        ])
    uuid = str(uuid4())
    generator = LabelGenerator([label]).add_url_to_masks(signer(uuid))
    assert label.annotations[0].value.mask.url != uuid
    assert next(generator).annotations[0].value.mask.url == uuid
    assert label.annotations[0].value.mask.url == uuid



from labelbox.data.metrics.iou.iou import miou_metric
from pytest_cases import parametrize, fixture_ref
from unittest.mock import patch
import math
import numpy as np
import base64

from labelbox.data.metrics.iou import data_row_miou, feature_miou_metric
from labelbox.data.serialization import NDJsonConverter, LBV1Converter
from labelbox.data.annotation_types import Label, ImageData, Mask


def check_iou(pair, mask=None):
    default = Label(data=ImageData(
        uid="ckppihxc10005aeyjen11h7jh", media_attributes=None, metadata=None))
    prediction = next(NDJsonConverter.deserialize(pair.predictions), default)
    label = next(LBV1Converter.deserialize([pair.labels]))
    if mask:
        for annotation in [*prediction.annotations, *label.annotations]:
            if isinstance(annotation.value, Mask):
                annotation.value.mask.arr = np.frombuffer(
                    base64.b64decode(annotation.value.mask.url.encode('utf-8')),
                    dtype=np.uint8).reshape((32, 32, 3))

    for include_subclasses, expected_attr_name in [[
            True, 'expected'
    ], [False, 'expected_without_subclasses']]:
        assert math.isclose(
            data_row_miou(label,
                          prediction,
                          include_subclasses=include_subclasses),
            getattr(pair, expected_attr_name))
        assert math.isclose(
            miou_metric(label.annotations,
                        prediction.annotations,
                        include_subclasses=include_subclasses)[0].value,
            getattr(pair, expected_attr_name))
        feature_ious = feature_miou_metric(
            label.annotations,
            prediction.annotations,
            include_subclasses=include_subclasses)
        assert len(
            feature_ious
        ) == 1  # The tests run here should only have one class present.
        assert math.isclose(feature_ious[0].value,
                            getattr(pair, expected_attr_name))


def check_iou_checklist(pair, mask=None):
    """specialized test since checklists have more than one feature ious """
    default = Label(data=ImageData(uid="ckppihxc10005aeyjen11h7jh"))
    prediction = next(NDJsonConverter.deserialize(pair.predictions), default)
    label = next(LBV1Converter.deserialize([pair.labels]))
    if mask:
        for annotation in [*prediction.annotations, *label.annotations]:
            if isinstance(annotation.value, Mask):
                annotation.value.mask.arr = np.frombuffer(
                    base64.b64decode(annotation.value.mask.url.encode('utf-8')),
                    dtype=np.uint8).reshape((32, 32, 3))
    assert math.isclose(data_row_miou(label, prediction),
                        pair.data_row_expected)
    assert math.isclose(
        miou_metric(label.annotations, prediction.annotations)[0].value,
        pair.data_row_expected)
    feature_ious = feature_miou_metric(label.annotations,
                                       prediction.annotations)
    mapping = {}
    for iou in feature_ious:
        if not mapping.get(iou.value, None):
            mapping[iou.value] = 0
        mapping[iou.value] += 1
    assert mapping == pair.expected


def strings_to_fixtures(strings):
    return [fixture_ref(x) for x in strings]


def test_overlapping(polygon_pair, box_pair, mask_pair):
    check_iou(polygon_pair)
    check_iou(box_pair)
    check_iou(mask_pair, True)


@parametrize("pair",
             strings_to_fixtures([
                 "unmatched_label",
                 "unmatched_prediction",
             ]))
def test_unmatched(pair):
    check_iou(pair)


@parametrize(
    "pair",
    strings_to_fixtures([
        "empty_radio_label",
        "matching_radio",
        "empty_radio_prediction",
    ]))
def test_radio(pair):
    check_iou(pair)


@parametrize(
    "pair",
    strings_to_fixtures([
        "matching_checklist",
        "partially_matching_checklist_1",
        "partially_matching_checklist_2",
        "partially_matching_checklist_3",
        "empty_checklist_label",
        "empty_checklist_prediction",
    ]))
def test_checklist(pair):
    check_iou_checklist(pair)


@parametrize("pair", strings_to_fixtures(["matching_text",
                                          "not_matching_text"]))
def test_text(pair):
    check_iou(pair)


@parametrize(
    "pair",
    strings_to_fixtures(
        ["test_box_with_wrong_subclass", "test_box_with_subclass"]))
def test_vector_with_subclass(pair):
    check_iou(pair)


@parametrize("pair", strings_to_fixtures(["point_pair", "line_pair"]))
def test_others(pair):
    check_iou(pair)


@parametrize(
    "pair",
    strings_to_fixtures(
        ["matching_ner", "no_matching_ner", "partial_matching_ner"]))
def test_ner(pair):
    check_iou(pair)



import labelbox.types as lb_types


def test_mask_frame():
    mask_frame = lb_types.MaskFrame(index=1,
                                    instance_uri="http://path/to/frame.png")
    assert mask_frame.dict(by_alias=True) == {
        'index': 1,
        'imBytes': None,
        'instanceURI': 'http://path/to/frame.png'
    }


def test_mask_instance():
    mask_instance = lb_types.MaskInstance(color_rgb=(0, 0, 255), name="mask1")
    assert mask_instance.dict(by_alias=True) == {
        'colorRGB': (0, 0, 255),
        'name': 'mask1'
    }



import json
import pytest


@pytest.fixture
def ndjson_content():
    line = """{"uuid": "9fd9a92e-2560-4e77-81d4-b2e955800092", "schemaId": "ckaeasyfk004y0y7wyye5epgu", "dataRow": {"id": "ck7kftpan8ir008910yf07r9c"}, "bbox": {"top": 48, "left": 58, "height": 865, "width": 1512}}
{"uuid": "29b878f3-c2b4-4dbf-9f22-a795f0720125", "schemaId": "ckapgvrl7007q0y7ujkjkaaxt", "dataRow": {"id": "ck7kftpan8ir008910yf07r9c"}, "polygon": [{"x": 147.692, "y": 118.154}, {"x": 142.769, "y": 404.923}, {"x": 57.846, "y": 318.769}, {"x": 28.308, "y": 169.846}]}"""
    expected_objects = [{
        'uuid': '9fd9a92e-2560-4e77-81d4-b2e955800092',
        'schemaId': 'ckaeasyfk004y0y7wyye5epgu',
        'dataRow': {
            'id': 'ck7kftpan8ir008910yf07r9c'
        },
        'bbox': {
            'top': 48,
            'left': 58,
            'height': 865,
            'width': 1512
        }
    }, {
        'uuid':
            '29b878f3-c2b4-4dbf-9f22-a795f0720125',
        'schemaId':
            'ckapgvrl7007q0y7ujkjkaaxt',
        'dataRow': {
            'id': 'ck7kftpan8ir008910yf07r9c'
        },
        'polygon': [{
            'x': 147.692,
            'y': 118.154
        }, {
            'x': 142.769,
            'y': 404.923
        }, {
            'x': 57.846,
            'y': 318.769
        }, {
            'x': 28.308,
            'y': 169.846
        }]
    }]

    return line, expected_objects


@pytest.fixture
def ndjson_content_with_nonascii_and_line_breaks():
    line = '{"id": "2489651127", "type": "PushEvent", "actor": {"id": 1459915, "login": "xtuaok", "gravatar_id": "", "url": "https://api.github.com/users/xtuaok", "avatar_url": "https://avatars.githubusercontent.com/u/1459915?"}, "repo": {"id": 6719841, "name": "xtuaok/twitter_track_following", "url": "https://api.github.com/repos/xtuaok/twitter_track_following"}, "payload": {"push_id": 536864008, "size": 1, "distinct_size": 1, "ref": "refs/heads/xtuaok", "head": "afb8afe306c7893d93d383a06e4d9df53b41bf47", "before": "4671b4868f1a060f2ed64d8268cd22d514a84e63", "commits": [{"sha": "afb8afe306c7893d93d383a06e4d9df53b41bf47", "author": {"email": "47cb89439b2d6961b59dff4298e837f67aa77389@gmail.com", "name": "Tomonori Tamagawa"}, "message": "Update ID 949438177,, - screen_name: chomado, - name: ちょまど@初詣おみくじ凶, - description: ( *ﾟ▽ﾟ* っ)З腐女子！絵描き！| H26新卒文系SE (入社して4ヶ月目の8月にSIer(適応障害になった)を辞職し開発者に転職) | H26秋応用情報合格！| 自作bot (in PHP) chomado_bot | プログラミングガチ初心者, - location:", "distinct": true, "url": "https://api.github.com/repos/xtuaok/twitter_track_following/commits/afb8afe306c7893d93d383a06e4d9df53b41bf47"}]}, "public": true, "created_at": "2015-01-01T15:00:10Z"}'
    expected_objects = [{
        'id': '2489651127',
        'type': 'PushEvent',
        'actor': {
            'id': 1459915,
            'login': 'xtuaok',
            'gravatar_id': '',
            'url': 'https://api.github.com/users/xtuaok',
            'avatar_url': 'https://avatars.githubusercontent.com/u/1459915?'
        },
        'repo': {
            'id': 6719841,
            'name': 'xtuaok/twitter_track_following',
            'url': 'https://api.github.com/repos/xtuaok/twitter_track_following'
        },
        'payload': {
            'push_id':
                536864008,
            'size':
                1,
            'distinct_size':
                1,
            'ref':
                'refs/heads/xtuaok',
            'head':
                'afb8afe306c7893d93d383a06e4d9df53b41bf47',
            'before':
                '4671b4868f1a060f2ed64d8268cd22d514a84e63',
            'commits': [{
                'sha':
                    'afb8afe306c7893d93d383a06e4d9df53b41bf47',
                'author': {
                    'email':
                        '47cb89439b2d6961b59dff4298e837f67aa77389@gmail.com',
                    'name':
                        'Tomonori Tamagawa'
                },
                'message':
                    'Update ID 949438177,, - screen_name: chomado, - name: ちょまど@初詣おみくじ凶, - description: ( *ﾟ▽ﾟ* っ)З腐女子！絵描き！| H26新卒文系SE (入社して4ヶ月目の8月にSIer(適応障害になった)を辞職し開発者に転職) | H26秋応用情報合格！| 自作bot (in PHP) chomado_bot | プログラミングガチ初心者, - location:',
                'distinct':
                    True,
                'url':
                    'https://api.github.com/repos/xtuaok/twitter_track_following/commits/afb8afe306c7893d93d383a06e4d9df53b41bf47'
            }]
        },
        'public': True,
        'created_at': '2015-01-01T15:00:10Z'
    }]
    return line, expected_objects


@pytest.fixture
def generate_random_ndjson(rand_gen):

    def _generate_random_ndjson(lines: int = 10):
        return [
            json.dumps({"data_row": {
                "id": rand_gen(str)
            }}) for _ in range(lines)
        ]

    return _generate_random_ndjson


@pytest.fixture
def mock_response():

    class MockResponse:

        def __init__(self, text: str, exception: Exception = None) -> None:
            self._text = text
            self._exception = exception

        @property
        def text(self):
            return self._text

        @property
        def content(self):
            return self._text

        def raise_for_status(self):
            if self._exception:
                raise self._exception

    return MockResponse



from labelbox.pydantic_compat import ValidationError
import numpy as np

import labelbox.types as lb_types
from labelbox import OntologyBuilder, Tool, Classification as OClassification, Option
from labelbox.data.annotation_types import (ClassificationAnswer, Radio, Text,
                                            ClassificationAnnotation,
                                            PromptText,
                                            ObjectAnnotation, Point, Line,
                                            ImageData, Label)
import pytest


def test_schema_assignment_geometry():
    name = "line_feature"
    label = Label(
        data=ImageData(arr=np.ones((32, 32, 3), dtype=np.uint8)),
        annotations=[
            ObjectAnnotation(
                value=Line(
                    points=[Point(x=1, y=2), Point(x=2, y=2)]),
                name=name,
            )
        ])
    feature_schema_id = "expected_id"
    ontology = OntologyBuilder(tools=[
        Tool(Tool.Type.LINE, name=name, feature_schema_id=feature_schema_id)
    ])
    label.assign_feature_schema_ids(ontology)

    assert label.annotations[0].feature_schema_id == feature_schema_id


def test_schema_assignment_classification():
    radio_name = "radio_name"
    text_name = "text_name"
    option_name = "my_option"

    label = Label(data=ImageData(arr=np.ones((32, 32, 3), dtype=np.uint8)),
                  annotations=[
                      ClassificationAnnotation(value=Radio(
                          answer=ClassificationAnswer(name=option_name)),
                                               name=radio_name),
                      ClassificationAnnotation(value=Text(answer="some text"),
                                               name=text_name)
                  ])
    radio_schema_id = "radio_schema_id"
    text_schema_id = "text_schema_id"
    option_schema_id = "option_schema_id"
    ontology = OntologyBuilder(
        tools=[],
        classifications=[
            OClassification(class_type=OClassification.Type.RADIO,
                            name=radio_name,
                            feature_schema_id=radio_schema_id,
                            options=[
                                Option(value=option_name,
                                       feature_schema_id=option_schema_id)
                            ]),
            OClassification(
                class_type=OClassification.Type.TEXT,
                name=text_name,
                feature_schema_id=text_schema_id,
            )
        ])
    label.assign_feature_schema_ids(ontology)
    assert label.annotations[0].feature_schema_id == radio_schema_id
    assert label.annotations[1].feature_schema_id == text_schema_id
    assert label.annotations[
        0].value.answer.feature_schema_id == option_schema_id


def test_schema_assignment_subclass():
    name = "line_feature"
    radio_name = "radio_name"
    option_name = "my_option"
    classification = ClassificationAnnotation(
        name=radio_name,
        value=Radio(answer=ClassificationAnswer(name=option_name)),
    )
    label = Label(
        data=ImageData(arr=np.ones((32, 32, 3), dtype=np.uint8)),
        annotations=[
            ObjectAnnotation(value=Line(
                points=[Point(x=1, y=2), Point(x=2, y=2)]),
                             name=name,
                             classifications=[classification])
        ])
    feature_schema_id = "expected_id"
    classification_schema_id = "classification_id"
    option_schema_id = "option_schema_id"
    ontology = OntologyBuilder(tools=[
        Tool(Tool.Type.LINE,
             name=name,
             feature_schema_id=feature_schema_id,
             classifications=[
                 OClassification(class_type=OClassification.Type.RADIO,
                                 name=radio_name,
                                 feature_schema_id=classification_schema_id,
                                 options=[
                                     Option(value=option_name,
                                            feature_schema_id=option_schema_id)
                                 ])
             ])
    ])
    label.assign_feature_schema_ids(ontology)
    assert label.annotations[0].feature_schema_id == feature_schema_id
    assert label.annotations[0].classifications[
        0].feature_schema_id == classification_schema_id
    assert label.annotations[0].classifications[
        0].value.answer.feature_schema_id == option_schema_id


def test_highly_nested():
    name = "line_feature"
    radio_name = "radio_name"
    nested_name = "nested_name"
    option_name = "my_option"
    nested_option_name = "nested_option_name"
    classification = ClassificationAnnotation(
        name=radio_name,
        value=Radio(answer=ClassificationAnswer(name=option_name)),
        classifications=[
            ClassificationAnnotation(value=Radio(answer=ClassificationAnswer(
                name=nested_option_name)),
                                     name=nested_name)
        ])
    label = Label(
        data=ImageData(arr=np.ones((32, 32, 3), dtype=np.uint8)),
        annotations=[
            ObjectAnnotation(value=Line(
                points=[Point(x=1, y=2), Point(x=2, y=2)]),
                             name=name,
                             classifications=[classification])
        ])
    feature_schema_id = "expected_id"
    classification_schema_id = "classification_id"
    nested_classification_schema_id = "nested_classification_schema_id"
    option_schema_id = "option_schema_id"
    ontology = OntologyBuilder(tools=[
        Tool(Tool.Type.LINE,
             name=name,
             feature_schema_id=feature_schema_id,
             classifications=[
                 OClassification(
                     class_type=OClassification.Type.RADIO,
                     name=radio_name,
                     feature_schema_id=classification_schema_id,
                     options=[
                         Option(value=option_name,
                                feature_schema_id=option_schema_id,
                                options=[
                                    OClassification(
                                        class_type=OClassification.Type.RADIO,
                                        name=nested_name,
                                        feature_schema_id=
                                        nested_classification_schema_id,
                                        options=[
                                            Option(
                                                value=nested_option_name,
                                                feature_schema_id=
                                                nested_classification_schema_id)
                                        ])
                                ])
                     ])
             ])
    ])
    label.assign_feature_schema_ids(ontology)
    assert label.annotations[0].feature_schema_id == feature_schema_id
    assert label.annotations[0].classifications[
        0].feature_schema_id == classification_schema_id
    assert label.annotations[0].classifications[
        0].value.answer.feature_schema_id == option_schema_id


def test_schema_assignment_confidence():
    name = "line_feature"
    label = Label(data=ImageData(arr=np.ones((32, 32, 3), dtype=np.uint8)),
                  annotations=[
                      ObjectAnnotation(value=Line(
                          points=[Point(x=1, y=2),
                                  Point(x=2, y=2)],),
                                       name=name,
                                       confidence=0.914)
                  ])

    assert label.annotations[0].confidence == 0.914


def test_initialize_label_no_coercion():
    global_key = 'global-key'
    ner_annotation = lb_types.ObjectAnnotation(
        name="ner",
        value=lb_types.ConversationEntity(start=0, end=8, message_id="4"))
    label = Label(data=lb_types.ConversationData(global_key=global_key),
                  annotations=[ner_annotation])
    assert isinstance(label.data, lb_types.ConversationData)
    assert label.data.global_key == global_key

def test_prompt_classification_validation():
    global_key = 'global-key'
    prompt_text = lb_types.PromptClassificationAnnotation(
        name="prompt text",
        value=PromptText(answer="test")
    )
    prompt_text_2 = lb_types.PromptClassificationAnnotation(
        name="prompt text",
        value=PromptText(answer="test")
    )
    with pytest.raises(ValidationError) as e_info:
        label = Label(data={"global_key": global_key},
                  annotations=[prompt_text, prompt_text_2])



import json

import pytest

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter


@pytest.mark.parametrize("file_path", [
    'tests/data/assets/labelbox_v1/highly_nested_image.json',
    'tests/data/assets/labelbox_v1/image_export.json'
])
#TODO: some checklists from the export come in as [checklist ans: []]
# while others are checklist ans: []... when we can figure out why we sometimes
# have extra brackets, we can look into testing nested checklist answers
# and ensuring the export's output matches deserialized/serialized output
def test_image(file_path):
    with open(file_path, 'r') as file:
        payload = json.load(file)

    collection = LBV1Converter.deserialize([payload])
    serialized = next(LBV1Converter.serialize(collection))

    # We are storing the media types now.
    payload['media_type'] = 'image'
    payload['Global Key'] = None

    assert serialized.keys() == payload.keys()

    for key in serialized:
        if key != 'Label':
            assert serialized[key] == payload[key]
        elif key == 'Label':
            for annotation_a, annotation_b in zip(serialized[key]['objects'],
                                                  payload[key]['objects']):
                annotation_b['page'] = None
                annotation_b['unit'] = None
                if not len(annotation_a['classifications']):
                    # We don't add a classification key to the payload if there is no classifications.
                    annotation_a.pop('classifications')

                if isinstance(annotation_b.get('classifications'),
                              list) and len(annotation_b['classifications']):
                    if isinstance(annotation_b['classifications'][0], list):
                        annotation_b['classifications'] = annotation_b[
                            'classifications'][0]
                assert annotation_a == annotation_b


# After check the nd serializer on this shit.. It should work for almost everything (except the other horse shit..)



import json

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter


def round_dict(data):
    for key in data:
        if isinstance(data[key], float):
            data[key] = int(data[key])
        elif isinstance(data[key], dict):
            data[key] = round_dict(data[key])
    return data


def test_video():
    payload = json.load(
        open('tests/data/assets/labelbox_v1/video_export.json', 'r'))
    collection = LBV1Converter.deserialize([payload])
    serialized = next(LBV1Converter.serialize(collection))
    payload['media_type'] = 'video'
    payload['Global Key'] = None
    assert serialized.keys() == payload.keys()
    for key in serialized:
        if key != 'Label':
            assert serialized[key] == payload[key]
        elif key == 'Label':
            for annotation_a, annotation_b in zip(serialized[key],
                                                  payload[key]):
                assert annotation_a['frameNumber'] == annotation_b[
                    'frameNumber']
                assert annotation_a['classifications'] == annotation_b[
                    'classifications']

                for obj_a, obj_b in zip(annotation_a['objects'],
                                        annotation_b['objects']):
                    obj_b['page'] = None
                    obj_b['unit'] = None
                    obj_a = round_dict(obj_a)
                    obj_b = round_dict(obj_b)
                    assert obj_a == obj_b



import pytest

from labelbox.exceptions import error_message_for_unparsed_graphql_error


@pytest.mark.parametrize('exception_message, expected_result', [
    ("Unparsed errors on query execution: [{'message': 'Cannot create model config for project because model setup is complete'}]",
     "Cannot create model config for project because model setup is complete"),
    ("blah blah blah", "Unknown error"),
])
def test_client_unparsed_exception_messages(exception_message, expected_result):
    assert error_message_for_unparsed_graphql_error(
        exception_message) == expected_result



from typing import Dict, Any

from labelbox import pydantic_compat


class TextEntity(pydantic_compat.BaseModel):
    """ Represents a text entity """
    start: int
    end: int
    extra: Dict[str, Any] = {}

    @pydantic_compat.root_validator
    def validate_start_end(cls, values):
        if 'start' in values and 'end' in values:
            if (isinstance(values['start'], int) and
                    values['start'] > values['end']):
                raise ValueError(
                    "Location end must be greater or equal to start")
        return values



from typing import Optional, List

from labelbox import pydantic_compat

from labelbox.exceptions import ConfidenceNotSupportedException, CustomMetricsNotSupportedException


class ConfidenceMixin(pydantic_compat.BaseModel):
    confidence: Optional[float] = None

    @pydantic_compat.validator("confidence")
    def confidence_valid_float(cls, value):
        if value is None:
            return value
        if not isinstance(value, (int, float)) or not 0 <= value <= 1:
            raise ValueError("must be a number within [0,1] range")
        return value

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)
        if "confidence" in res and res["confidence"] is None:
            res.pop("confidence")
        return res


class ConfidenceNotSupportedMixin:

    def __new__(cls, *args, **kwargs):
        if "confidence" in kwargs:
            raise ConfidenceNotSupportedException(
                "Confidence is not supported for this annotation type yet")
        return super().__new__(cls)


class CustomMetric(pydantic_compat.BaseModel):
    name: str
    value: float

    @pydantic_compat.validator("name")
    def confidence_valid_float(cls, value):
        if not isinstance(value, str):
            raise ValueError("Name must be a string")
        return value

    @pydantic_compat.validator("value")
    def value_valid_float(cls, value):
        if not isinstance(value, (int, float)):
            raise ValueError("Value must be a number")
        return value


class CustomMetricsMixin(pydantic_compat.BaseModel):
    custom_metrics: Optional[List[CustomMetric]] = None

    def dict(self, *args, **kwargs):
        res = super().dict(*args, **kwargs)

        if "customMetrics" in res and res["customMetrics"] is None:
            res.pop("customMetrics")

        if "custom_metrics" in res and res["custom_metrics"] is None:
            res.pop("custom_metrics")

        return res


class CustomMetricsNotSupportedMixin:

    def __new__(cls, *args, **kwargs):
        if "custom_metrics" in kwargs:
            raise CustomMetricsNotSupportedException(
                "Custom metrics is not supported for this annotation type yet")
        return super().__new__(cls)



from itertools import chain
from typing import Any, Dict

from labelbox import utils
from labelbox.exceptions import InvalidQueryError, InvalidAttributeError, MalformedQueryException
from labelbox.orm.comparison import LogicalExpression, Comparison
from labelbox.orm.model import Field, Relationship, Entity
""" Common query creation functionality. """


def format_param_declaration(params):
    """ Formats the parameters dictionary into a declaration of GraphQL
    query parameters.

    Args:
        params (dict): keys are query param names and values are
            (value, (field|relationship)) tuples.
    Return:
        str, the declaration of query parameters.
    """
    if not params:
        return ""

    def attr_type(attr):
        if isinstance(attr, Field):
            return attr.field_type.name
        else:
            return Field.Type.ID.name

    return "(" + ", ".join("$%s: %s!" % (param, attr_type(attr))
                           for param, (_, attr) in params.items()) + ")"


def results_query_part(entity):
    """ Generates the results part of the query. The results contain
    all the entity's fields as well as prefetched relationships.

    Note that this is a recursive function. If there is a cycle in the
    prefetched relationship graph, this function will recurse infinitely.

    Args:
        entity (type): The entity which needs fetching.
    """
    # Query for fields
    fields = [
        field.result_subquery
        if field.result_subquery is not None else field.graphql_name
        for field in entity.fields()
    ]

    # Query for cached relationships
    fields.extend([
        Query(rel.graphql_name, rel.destination_type).format()[0]
        for rel in entity.relationships()
        if rel.cache
    ])
    return " ".join(fields)


class Query:
    """ A data structure used during the construction of a query. Supports
    subquery (also Query object) nesting for relationship. """

    def __init__(self,
                 what,
                 subquery,
                 where=None,
                 paginate=False,
                 order_by=None):
        """ Initializer.
        Args:
            what (str): What is being queried. Typically an object type in
                singular or plural (i.e. "project" or "projects").
            subquery (Query or type): Either a Query object that is formatted
                recursively or a Entity subtype in which case the standard
                results (see `results_query_part`) are retrieved.
            where (None, Comparison or LogicalExpression): the filtering clause.
            paginate (bool): If the "%skip %first" pagination substring should
                be added to the query. Used for collection pagination in combination
                with PaginatedCollection.
            order_by (tuple): A tuple consisting of (Field, Field.Order) indicating
                how the query should sort the collection.
        """
        self.what = what
        self.subquery = subquery
        self.paginate = paginate
        self.where = where
        self.order_by = order_by

    def format_subquery(self):
        """ Formats the subquery (a Query or Entity subtype). """
        if isinstance(self.subquery, Query):
            return self.subquery.format()
        elif issubclass(self.subquery, Entity):
            return results_query_part(self.subquery), {}
        else:
            raise MalformedQueryException()

    def format_clauses(self, params):
        """ Formats the where, order_by and pagination clauses.
        Args:
            params (dict): The current parameter dictionary.
        """

        def format_where(node):
            """ Helper that resursively constructs a where clause from a
            LogicalExpression tree (leaf nodes are Comparisons). """
            COMPARISON_TO_SUFFIX = {
                Comparison.Op.EQ: "",
                Comparison.Op.NE: "_not",
                Comparison.Op.LT: "_lt",
                Comparison.Op.GT: "_gt",
                Comparison.Op.LE: "_lte",
                Comparison.Op.GE: "_gte",
            }
            assert isinstance(node, (Comparison, LogicalExpression))
            if isinstance(node, Comparison):
                param_name = "param_%d" % len(params)
                params[param_name] = (node.value, node.field)
                return "{%s%s: $%s}" % (node.field.graphql_name,
                                        COMPARISON_TO_SUFFIX[node.op],
                                        param_name)
            if node.op == LogicalExpression.Op.NOT:
                return "{NOT: [%s]}" % format_where(node.first)

            return "{%s: [%s, %s]}" % (node.op.name.upper(),
                                       format_where(node.first),
                                       format_where(node.second))

        paginate = "skip: %d first: %d" if self.paginate else ""

        where = "where: %s" % format_where(self.where) if self.where else ""

        if self.order_by:
            order_by = "orderBy: %s_%s" % (self.order_by[0].graphql_name,
                                           self.order_by[1].name.upper())
        else:
            order_by = ""

        clauses = " ".join(filter(None, (where, paginate, order_by)))
        return "(" + clauses + ")" if clauses else ""

    def format(self):
        """ Formats the full query but without "query" prefix, query name
        and parameter declaration.
        Return:
            (str, dict) tuple. str is the query and dict maps parameter
            names to (value, field) tuples.
        """
        subquery, params = self.format_subquery()
        clauses = self.format_clauses(params)
        query = "%s%s{%s}" % (self.what, clauses, subquery)
        return query, params

    def format_top(self, name):
        """ Formats the full query including "query" prefix, query name
        and parameter declaration. The result of this function can be
        sent to the Client object for execution.

        Args:
            name (str): Query name, without the "PyApi" suffix, it's appended
                automatically by this method.
        Return:
            (str, dict) tuple. str is the full query and dict maps parameter
                names to parameter values.
        """
        query, params = self.format()
        param_declaration = format_param_declaration(params)
        query = "query %sPyApi%s{%s}" % (name, param_declaration, query)
        return query, {param: value for param, (value, _) in params.items()}


def get_single(entity, uid):
    """ Constructs the query and params dict for obtaining a single object. Either
    on ID, or without params.
    Args:
        entity (type): An Entity subtype being obtained.
        uid (str): The ID of the sought object. It can be None, which is legal for
            DB types that have a default object being returned (User and
            Organization).
    """
    type_name = entity.type_name()
    where = entity.uid == uid if uid else None
    return Query(utils.camel_case(type_name), entity,
                 where).format_top("Get" + type_name)


def logical_ops(where):
    """ Returns a generator that yields all the logical operator
    type objects (`LogicalExpression.Op` instances) from a where
    clause.

    Args:
        where (LogicalExpression, Comparison or None): The where
            clause used for filtering in a query.
    Return:
        See above.
    """
    if isinstance(where, LogicalExpression):
        yield where.op
        for f in chain(logical_ops(where.first), logical_ops(where.second)):
            yield f


def check_where_clause(entity, where):
    """ Checks the `where` clause of a query. A `where` clause is legal
    if it only refers to fields found in the entity it's defined for.
    Since only AND logical operations are supported server-side at the
    moment, logical OR and NOT are illegal.

    Args:
        entity (type): An Entity subclass type.
        where (LogicalExpression or Comparison): The `where` clause of
            query.
    Return:
        bool indicating if `where` is legal for `entity`.
    """

    def fields(where):
        """ Yields all the fields in a `where` clause. """
        if isinstance(where, LogicalExpression):
            for f in chain(fields(where.first), fields(where.second)):
                yield f
        elif isinstance(where, Comparison):
            yield where.field

    # The `deleted` field is a special case, ignore it.
    where_fields = [f for f in fields(where) if f != Entity.deleted]
    invalid_fields = set(where_fields) - set(entity.fields())
    if invalid_fields:
        raise InvalidAttributeError(entity, invalid_fields)

    if len(set(where_fields)) != len(where_fields):
        raise InvalidQueryError(
            "Where clause contains multiple comparisons for "
            "the same field: %r." % where)

    if set(logical_ops(where)) not in (set(), {LogicalExpression.Op.AND}):
        raise InvalidQueryError("Currently only AND logical ops are allowed in "
                                "the where clause of a query.")


def check_order_by_clause(entity, order_by):
    """ Checks that the `order_by` clause field is a part of `entity`.

    Args:
        entity (type): An Entity subclass type.
        order_by ((field, ordering)): The ordering tuple consisting of
            a field and sort ordering (ascending or descending).
    Return:
        bool indicating if `order_by` is legal for `entity`.
    """
    if order_by is not None:
        field, _ = order_by
        if field not in entity.fields():
            raise InvalidAttributeError(entity, field)


def get_all(entity, where):
    """ Constructs a query that fetches all items of the given type. The
    resulting query is intended to be used for pagination, it contains
    two python-string int-placeholders (%d) for 'skip' and 'first'
    pagination parameters.

    Args:
        entity (type): The object type being queried.
        where (Comparison, LogicalExpression or None): The `where` clause
            for filtering.
    Return:
        (str, dict) tuple that is the query string and parameters.
    """
    check_where_clause(entity, where)
    type_name = entity.type_name()
    query = Query(utils.camel_case(type_name) + "s", entity, where, True)
    return query.format_top("Get" + type_name + "s")


def relationship(source, relationship, where, order_by):
    """ Constructs a query that fetches all items from a -to-many
    relationship. To be used like:
        >>> project = ...
        >>> query_str, params = relationship(Project, "datasets", Dataset)
        >>> datasets = PaginatedCollection(
            client, query_str, params, ["project", "datasets"],
            Dataset)

    The resulting query is intended to be used for pagination, it contains
    two python-string int-placeholders (%d) for 'skip' and 'first'
    pagination parameters.

    Args:
        source (DbObject or type): If a `DbObject` then the source of the
            relationship (the query originates from that particular object).
            If `type`, then the source of the relationship is implicit, even
            without the ID. Used for expanding from Organization.
        relationship (Relationship): The relationship.
        where (Comparison, LogicalExpression or None): The `where` clause
            for filtering.
        order_by (None or (Field, Field.Order): The `order_by` clause for
            sorting results.
    Return:
        (str, dict) tuple that is the query string and parameters.
    """
    check_where_clause(relationship.destination_type, where)
    check_order_by_clause(relationship.destination_type, order_by)
    to_many = relationship.relationship_type == Relationship.Type.ToMany
    subquery = Query(relationship.graphql_name, relationship.destination_type,
                     where, to_many, order_by)
    query_where = type(source).uid == source.uid if isinstance(source, Entity) \
        else None
    query = Query(utils.camel_case(source.type_name()), subquery, query_where)
    return query.format_top("Get" + source.type_name() +
                            utils.title_case(relationship.graphql_name))


def create(entity, data):
    """ Generates a query and parameters for creating a new DB object.

    Args:
        entity (type): An Entity subtype indicating which kind of
            DB object needs to be created.
        data (dict): A dict that maps Fields and Relationships to values, new
            object data.
    Return:
        (query_string, parameters)
    """
    type_name = entity.type_name()

    def format_param_value(attribute, param):
        if isinstance(attribute, Field):
            return "%s: $%s" % (attribute.graphql_name, param)
        else:
            return "%s: {connect: {id: $%s}}" % (utils.camel_case(
                attribute.graphql_name), param)

    # Convert data to params
    params = {
        field.graphql_name: (value, field) for field, value in data.items()
    }

    query_str = """mutation Create%sPyApi%s{create%s(data: {%s}) {%s}} """ % (
        type_name, format_param_declaration(params), type_name, " ".join(
            format_param_value(attribute, param)
            for param, (_, attribute) in params.items()),
        results_query_part(entity))

    return query_str, {name: value for name, (value, _) in params.items()}


def update_relationship(a, b, relationship, update):
    """ Updates the relationship in DB object `a` to connect or disconnect
    DB object `b`.

    Args:
        a (DbObject): The object being updated.
        b (DbObject): Object on the other side of the relationship.
        relationship (Relationship): The relationship from `a` to `b`.
        update (str): The type of update. Must be either `connect` or
            `disconnect`.
    Return:
        (query_string, query_parameters)
    """
    to_one_disconnect = update == "disconnect" and \
        relationship.relationship_type == Relationship.Type.ToOne

    a_uid_param = utils.camel_case(type(a).type_name()) + "Id"

    if not to_one_disconnect:
        b_uid_param = utils.camel_case(type(b).type_name()) + "Id"
        param_declr = "($%s: ID!, $%s: ID!)" % (a_uid_param, b_uid_param)
        b_query = "{id: $%s}" % b_uid_param
    else:
        param_declr = "($%s: ID!)" % a_uid_param
        b_query = "true"

    query_str = """mutation %s%sAnd%sPyApi%s{update%s(
        where: {id: $%s} data: {%s: {%s: %s}}) {id}} """ % (
        utils.title_case(update), type(a).type_name(), type(b).type_name(),
        param_declr, utils.title_case(type(a).type_name()), a_uid_param,
        relationship.graphql_name, update, b_query)

    if to_one_disconnect:
        params = {a_uid_param: a.uid}
    else:
        params = {a_uid_param: a.uid, b_uid_param: b.uid}

    return query_str, params


def update_fields(db_object, values):
    """ Creates a query that updates `db_object` fields with the
    given values.

    Args:
        db_object (DbObject): The DB object being updated.
        values (dict): Maps Fields to new values. All Fields
            must be legit fields in `db_object`.
    Return:
        (query_string, query_parameters)
    """
    type_name = db_object.type_name()
    id_param = "%sId" % type_name
    values_str = " ".join("%s: $%s" % (field.graphql_name, field.graphql_name)
                          for field, _ in values.items())
    params = {
        field.graphql_name: (value, field) for field, value in values.items()
    }
    params[id_param] = (db_object.uid, Entity.uid)

    query_str = """mutation update%sPyApi%s{update%s(
        where: {id: $%s} data: {%s}) {%s}} """ % (
        utils.title_case(type_name), format_param_declaration(params),
        type_name, id_param, values_str, results_query_part(type(db_object)))

    return query_str, {name: value for name, (value, _) in params.items()}


def delete(db_object):
    """ Generates a query that deletes the given `db_object` from the DB.

    Args:
        db_object (DbObject): The DB object being deleted.
    """
    id_param = "%sId" % db_object.type_name()
    query_str = """mutation delete%sPyApi%s{update%s(
        where: {id: $%s} data: {deleted: true}) {id}} """ % (
        db_object.type_name(), "($%s: ID!)" % id_param, db_object.type_name(),
        id_param)

    return query_str, {id_param: db_object.uid}


def bulk_delete(db_objects, use_where_clause):
    """ Generates a query that bulk-deletes the given `db_objects` from the
    DB.

    Args:
        db_objects (list): A list of DB objects of the same type.
        use_where_clause (bool): If the object IDs should be passed to the
            mutation in a `where` clause or directly as a mutation value.
    """
    type_name = db_objects[0].type_name()
    if use_where_clause:
        query_str = "mutation delete%ssPyApi{delete%ss(where: {%sIds: [%s]}){id}}"
    else:
        query_str = "mutation delete%ssPyApi{delete%ss(%sIds: [%s]){id}}"
    query_str = query_str % (
        utils.title_case(type_name), utils.title_case(type_name),
        utils.camel_case(type_name), ", ".join(
            '"%s"' % db_object.uid for db_object in db_objects))
    return query_str, {}


def where_as_dict(entity, node: Comparison) -> Dict[str, Any]:
    check_where_clause(entity, node)
    """ Helper that constructs a where clause from a Comparison node. """
    COMPARISON_TO_SUFFIX = {
        Comparison.Op.EQ: "",
        Comparison.Op.NE: "_not",
        Comparison.Op.LT: "_lt",
        Comparison.Op.GT: "_gt",
        Comparison.Op.LE: "_lte",
        Comparison.Op.GE: "_gte",
    }

    key = f"{node.field.graphql_name}{COMPARISON_TO_SUFFIX[node.op]}"
    return {key: node.value}


def order_by_as_string(entity, node: tuple) -> str:
    check_order_by_clause(entity, node)
    return f"{node[0].graphql_name}_{node[1].name.upper()}"



import uuid
import pytest
from labelbox import parser

from labelbox.schema.annotation_import import AnnotationImportState, LabelImport
"""
- Here we only want to check that the uploads are calling the validation
- Then with unit tests we can check the types of errors raised

"""


def test_create_with_url_arg(client, module_project,
                             annotation_import_test_helpers):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create(
        client=client,
        id=module_project.uid,
        name=name,
        url=url)
    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


def test_create_from_url(client, module_project,
                         annotation_import_test_helpers):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create_from_url(
        client=client,
        project_id=module_project.uid,
        name=name,
        url=url)
    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


def test_create_with_labels_arg(client, module_project, object_predictions,
                                annotation_import_test_helpers):
    """this test should check running state only to validate running, not completed"""
    name = str(uuid.uuid4())

    label_import = LabelImport.create(client=client,
                                      id=module_project.uid,
                                      name=name,
                                      labels=object_predictions)

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions)


def test_create_from_objects(client, module_project, object_predictions,
                             annotation_import_test_helpers):
    """this test should check running state only to validate running, not completed"""
    name = str(uuid.uuid4())

    label_import = LabelImport.create_from_objects(
        client=client,
        project_id=module_project.uid,
        name=name,
        labels=object_predictions)

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions)


def test_create_with_path_arg(client, tmp_path, module_project, object_predictions,
                              annotation_import_test_helpers):
    project = module_project
    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(object_predictions, f)

    label_import = LabelImport.create(client=client,
                                      id=project.uid,
                                      name=name,
                                      path=str(file_path))

    assert label_import.parent_id == project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions)


def test_create_from_local_file(client, tmp_path, module_project, object_predictions,
                                annotation_import_test_helpers):
    project = module_project
    name = str(uuid.uuid4())
    file_name = f"{name}.ndjson"
    file_path = tmp_path / file_name
    with file_path.open("w") as f:
        parser.dump(object_predictions, f)

    label_import = LabelImport.create_from_file(client=client,
                                               project_id=project.uid,
                                               name=name,
                                               path=str(file_path))

    assert label_import.parent_id == project.uid
    annotation_import_test_helpers.check_running_state(label_import, name)
    annotation_import_test_helpers.assert_file_content(
        label_import.input_file_url, object_predictions)


def test_get(client, module_project,
             annotation_import_test_helpers):
    name = str(uuid.uuid4())
    url = "https://storage.googleapis.com/labelbox-public-bucket/predictions_test_v2.ndjson"
    label_import = LabelImport.create_from_url(
        client=client,
        project_id=module_project.uid,
        name=name,
        url=url)

    assert label_import.parent_id == module_project.uid
    annotation_import_test_helpers.check_running_state(label_import, name, url)


@pytest.mark.slow
def test_wait_till_done(client, module_project, predictions):
    name = str(uuid.uuid4())
    label_import = LabelImport.create_from_objects(
        client=client,
        project_id=module_project.uid,
        name=name,
        labels=predictions)

    assert len(label_import.inputs) == len(predictions)
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.inputs) == len(predictions)
    input_uuids = [input_annot['uuid'] for input_annot in label_import.inputs]
    inference_uuids = [pred['uuid'] for pred in predictions]
    assert set(input_uuids) == set(inference_uuids)
    assert len(label_import.statuses) == len(predictions)
    status_uuids = [
        input_annot['uuid'] for input_annot in label_import.statuses
    ]
    assert set(input_uuids) == set(status_uuids)



from multiprocessing.dummy import Pool
import os
import time
import pytest
from google.api_core.exceptions import RetryError

from labelbox import Project, Dataset, User
import labelbox.client
import labelbox.exceptions


def test_missing_api_key():
    key = os.environ.get(labelbox.client._LABELBOX_API_KEY, None)
    if key is not None:
        del os.environ[labelbox.client._LABELBOX_API_KEY]

    with pytest.raises(labelbox.exceptions.AuthenticationError) as excinfo:
        labelbox.client.Client()

    assert excinfo.value.message == "Labelbox API key not provided"

    if key is not None:
        os.environ[labelbox.client._LABELBOX_API_KEY] = key


def test_bad_key(rand_gen):
    bad_key = "BAD_KEY_" + rand_gen(str)
    client = labelbox.client.Client(api_key=bad_key)

    with pytest.raises(labelbox.exceptions.AuthenticationError) as excinfo:
        client.create_project(name=rand_gen(str))


def test_syntax_error(client):
    with pytest.raises(labelbox.exceptions.InvalidQueryError) as excinfo:
        client.execute("asda", check_naming=False)
    assert excinfo.value.message.startswith("Syntax Error:")


def test_semantic_error(client):
    with pytest.raises(labelbox.exceptions.InvalidQueryError) as excinfo:
        client.execute("query {bbb {id}}", check_naming=False)
    assert excinfo.value.message.startswith("Cannot query field \"bbb\"")


def test_timeout_error(client, project):
    with pytest.raises(RetryError) as excinfo:
        query_str = """query getOntology {
        project (where: {id: %s}) {
            ontology {
                normalized
                }
            }
        } """ % (project.uid)

        # Setting connect timeout to 30s, and read timeout to 0.01s
        client.execute(query_str, check_naming=False, timeout=(30.0, 0.01))


def test_query_complexity_error(client):
    with pytest.raises(labelbox.exceptions.ValidationFailedError) as excinfo:
        client.execute("{projects {datasets {dataRows {labels {id}}}}}",
                       check_naming=False)
    assert excinfo.value.message == "Query complexity limit exceeded"


def test_resource_not_found_error(client):
    with pytest.raises(labelbox.exceptions.ResourceNotFoundError):
        client.get_project("invalid project ID")


def test_network_error(client):
    client = labelbox.client.Client(api_key=client.api_key,
                                    endpoint="not_a_valid_URL")

    with pytest.raises(labelbox.exceptions.NetworkError) as excinfo:
        client.create_project(name="Project name")


def test_invalid_attribute_error(
    client,
    rand_gen,
):
    # Creation
    with pytest.raises(labelbox.exceptions.InvalidAttributeError) as excinfo:
        client.create_project(name="Name", invalid_field="Whatever")
    assert excinfo.value.db_object_type == Project
    assert excinfo.value.field == "invalid_field"

    # Update
    project = client.create_project(name=rand_gen(str))
    with pytest.raises(labelbox.exceptions.InvalidAttributeError) as excinfo:
        project.update(invalid_field="Whatever")
    assert excinfo.value.db_object_type == Project
    assert excinfo.value.field == "invalid_field"

    # Top-level-fetch
    with pytest.raises(labelbox.exceptions.InvalidAttributeError) as excinfo:
        client.get_projects(where=User.email == "email")
    assert excinfo.value.db_object_type == Project
    assert excinfo.value.field == {User.email}


@pytest.mark.skip("timeouts cause failure before rate limit")
def test_api_limit_error(client):

    def get(arg):
        try:
            return client.get_user()
        except labelbox.exceptions.ApiLimitError as e:
            return e

    # Rate limited at 1500 + buffer
    n = 1600
    # max of 30 concurrency before the service becomes unavailable
    with Pool(30) as pool:
        start = time.time()
        results = list(pool.imap(get, range(n)), total=n)
        elapsed = time.time() - start

    assert elapsed < 60, "Didn't finish fast enough"
    assert labelbox.exceptions.ApiLimitError in {type(r) for r in results}

    # Sleep at the end of this test to allow other tests to execute.
    time.sleep(60)



import pytest


@pytest.mark.skip(
    'Skipping until we have a way to create slices programatically')
def test_export_v2_slice(client):
    # Since we don't have CRUD for slices, we'll just use the one that's already there
    SLICE_ID = "clk04g1e4000ryb0rgsvy1dty"
    slice = client.get_catalog_slice(SLICE_ID)
    task = slice.export_v2(params={
        "performance_details": False,
        "label_details": True
    })
    task.wait_till_done()
    assert task.status == "COMPLETE"
    assert task.errors is None
    assert len(task.result) != 0



import json

import pytest

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


@pytest.mark.parametrize("filename", [
    "tests/data/assets/ndjson/text_entity_import.json",
    "tests/data/assets/ndjson/text_entity_without_confidence_import.json"
])
def test_text_entity_import(filename: str):
    with open(filename, 'r') as file:
        data = json.load(file)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data



from typing import Dict, Any, Optional, Union, Tuple
from abc import ABC, abstractmethod

import geojson
import numpy as np
from labelbox import pydantic_compat

from shapely import geometry as geom


class Geometry(pydantic_compat.BaseModel, ABC):
    """Abstract base class for geometry objects
    """
    extra: Dict[str, Any] = {}

    @property
    def shapely(
        self
    ) -> Union[geom.Point, geom.LineString, geom.Polygon, geom.MultiPoint,
               geom.MultiLineString, geom.MultiPolygon]:
        return geom.shape(self.geometry)

    def get_or_create_canvas(self, height: Optional[int], width: Optional[int],
                             canvas: Optional[np.ndarray]) -> np.ndarray:
        if canvas is None:
            if height is None or width is None:
                raise ValueError(
                    "Must either provide canvas or height and width")
            canvas = np.zeros((height, width, 3), dtype=np.uint8)
        canvas = np.ascontiguousarray(canvas)
        return canvas

    @property
    @abstractmethod
    def geometry(self) -> geojson:
        pass

    @abstractmethod
    def draw(self,
             height: Optional[int] = None,
             width: Optional[int] = None,
             canvas: Optional[np.ndarray] = None,
             color: Optional[Union[int, Tuple[int, int, int]]] = None,
             thickness: Optional[int] = 1) -> np.ndarray:
        pass



import uuid
import pytest
from unittest.mock import MagicMock, patch

from labelbox.schema.annotation_import import MALPredictionImport, logger


def test_should_warn_user_about_unsupported_confidence():
    """this test should check running state only to validate running, not completed"""
    id = str(uuid.uuid4())

    labels = [
        {
            "bbox": {
                "height": 428,
                "left": 2089,
                "top": 1251,
                "width": 158
            },
            "classifications": [{
                "answer": [{
                    "schemaId": "ckrb1sfl8099e0y919v260awv",
                    "confidence": 0.894
                }],
                "schemaId": "ckrb1sfkn099c0y910wbo0p1a"
            }],
            "dataRow": {
                "id": "ckrb1sf1i1g7i0ybcdc6oc8ct"
            },
            "schemaId": "ckrb1sfjx099a0y914hl319ie",
            "uuid": "d009925d-91a3-4f67-abd9-753453f5a584"
        },
    ]
    with patch.object(MALPredictionImport, '_create_mal_import_from_bytes'):
        with patch.object(logger, 'warning') as warning_mock:
            MALPredictionImport.create_from_objects(client=MagicMock(),
                                                    project_id=id,
                                                    name=id,
                                                    predictions=labels)
            warning_mock.assert_called_once()
            "Confidence scores are not supported in MAL Prediction Import" in warning_mock.call_args_list[
                0].args[0]


def test_invalid_labels_format():
    """this test should confirm that annotations are required to be in a form of list"""
    id = str(uuid.uuid4())

    label = {
        "bbox": {
            "height": 428,
            "left": 2089,
            "top": 1251,
            "width": 158
        },
        "classifications": [{
            "answer": [{
                "schemaId": "ckrb1sfl8099e0y919v260awv",
                "confidence": 0.894
            }],
            "schemaId": "ckrb1sfkn099c0y910wbo0p1a"
        }],
        "dataRow": {
            "id": "ckrb1sf1i1g7i0ybcdc6oc8ct"
        },
        "schemaId": "ckrb1sfjx099a0y914hl319ie",
        "uuid": "3a83db52-75e0-49af-a171-234ce604502a"
    }

    with patch.object(MALPredictionImport, '_create_mal_import_from_bytes'):
        with pytest.raises(TypeError):
            MALPredictionImport.create_from_objects(client=MagicMock(),
                                                    project_id=id,
                                                    name=id,
                                                    predictions=label)



import pytest
from unittest.mock import patch

from labelbox import MediaType
from labelbox.schema.ontology_kind import OntologyKind
from labelbox.exceptions import MalformedQueryException


def test_create_chat_evaluation_ontology_project(
        client, chat_evaluation_ontology,
        live_chat_evaluation_project_with_new_dataset,
        offline_conversational_data_row, rand_gen):
    ontology = chat_evaluation_ontology

    # here we are essentially testing the ontology creation which is a fixture
    assert ontology
    assert ontology.name
    assert len(ontology.tools()) == 3
    for tool in ontology.tools():
        assert tool.schema_id
        assert tool.feature_schema_id

    assert (len(ontology.classifications()) == 6)
    for classification in ontology.classifications():
        assert classification.schema_id
        assert classification.feature_schema_id

    project = live_chat_evaluation_project_with_new_dataset
    assert project.model_setup_complete is None

    project.connect_ontology(ontology)

    assert project.labeling_frontend().name == "Editor"
    assert project.ontology().name == ontology.name

    with pytest.raises(
            ValueError,
            match="Cannot create batches for auto data generation projects"):
        project.create_batch(
            rand_gen(str),
            [offline_conversational_data_row.uid],  # sample of data row objects
        )

    with pytest.raises(
            ValueError,
            match="Cannot create batches for auto data generation projects"):
        with patch('labelbox.schema.project.MAX_SYNC_BATCH_ROW_COUNT',
                   new=0):  # force to async

            project.create_batch(
                rand_gen(str),
                [offline_conversational_data_row.uid
                ],  # sample of data row objects
            )


def test_create_chat_evaluation_ontology_project_existing_dataset(
        client, chat_evaluation_ontology,
        chat_evaluation_project_append_to_dataset):
    ontology = chat_evaluation_ontology

    project = chat_evaluation_project_append_to_dataset
    assert project
    project.connect_ontology(ontology)

    assert project.labeling_frontend().name == "Editor"
    assert project.ontology().name == ontology.name


@pytest.fixture
def tools_json():
    tools = [{
        'tool': 'message-single-selection',
        'name': 'model output single selection',
        'required': False,
        'color': '#ff0000',
        'classifications': [],
        'schemaNodeId': None,
        'featureSchemaId': None
    }, {
        'tool': 'message-multi-selection',
        'name': 'model output multi selection',
        'required': False,
        'color': '#00ff00',
        'classifications': [],
        'schemaNodeId': None,
        'featureSchemaId': None
    }, {
        'tool': 'message-ranking',
        'name': 'model output multi ranking',
        'required': False,
        'color': '#0000ff',
        'classifications': [],
        'schemaNodeId': None,
        'featureSchemaId': None
    }]

    return tools


@pytest.fixture
def features_from_json(client, tools_json):
    tools = tools_json
    features = {client.create_feature_schema(t) for t in tools if t}

    yield features

    for f in features:
        client.delete_unused_feature_schema(f.uid)


@pytest.fixture
def ontology_from_feature_ids(client, features_from_json):
    feature_ids = {f.uid for f in features_from_json}
    ontology = client.create_ontology_from_feature_schemas(
        name="test-model-chat-evaluation-ontology{rand_gen(str)}",
        feature_schema_ids=feature_ids,
        media_type=MediaType.Conversational,
        ontology_kind=OntologyKind.ModelEvaluation,
    )

    yield ontology

    client.delete_unused_ontology(ontology.uid)


def test_ontology_create_feature_schema(ontology_from_feature_ids,
                                        features_from_json, tools_json):
    created_ontology = ontology_from_feature_ids
    feature_schema_ids = {f.uid for f in features_from_json}
    tools_normalized = created_ontology.normalized['tools']
    tools = tools_json

    for tool in tools:
        generated_tool = next(
            t for t in tools_normalized if t['name'] == tool['name'])
        assert generated_tool['schemaNodeId'] is not None
        assert generated_tool['featureSchemaId'] in feature_schema_ids
        assert generated_tool['tool'] == tool['tool']
        assert generated_tool['name'] == tool['name']
        assert generated_tool['required'] == tool['required']
        assert generated_tool['color'] == tool['color']



import json
from typing import Dict, Any

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter

IGNORE_KEYS = [
    "Data Split", "media_type", "DataRow Metadata", "Media Attributes"
]


def round_dict(data: Dict[str, Any]) -> Dict[str, Any]:
    for key in data:
        if isinstance(data[key], float):
            data[key] = int(data[key])
        elif isinstance(data[key], dict):
            data[key] = round_dict(data[key])
    return data


def test_pdf():
    """
    Tests an export from a pdf document with only bounding boxes
    """
    payload = json.load(
        open('tests/data/assets/labelbox_v1/pdf_export.json', 'r'))
    collection = LBV1Converter.deserialize(payload)
    serialized = next(LBV1Converter.serialize(collection))

    payload = payload[0]  # only one document in the export

    serialized = {k: v for k, v in serialized.items() if k not in IGNORE_KEYS}

    assert serialized.keys() == payload.keys()
    for key in payload.keys():
        if key == 'Label':
            serialized_no_classes = [{
                k: v for k, v in dic.items() if k != 'classifications'
            } for dic in serialized[key]['objects']]
            serialized_round = [
                round_dict(dic) for dic in serialized_no_classes
            ]
            payload_round = [round_dict(dic) for dic in payload[key]['objects']]
            assert payload_round == serialized_round
        else:
            assert serialized[key] == payload[key]



from typing import Generator, TYPE_CHECKING

from labelbox.orm.db_object import DbObject, experimental
from labelbox.orm import query
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.exceptions import LabelboxError, ResourceNotFoundError
from io import StringIO
from labelbox import parser
import requests
import logging
import time
import warnings

if TYPE_CHECKING:
    from labelbox import Project

logger = logging.getLogger(__name__)


class Batch(DbObject):
    """ A Batch is a group of data rows submitted to a project for labeling

    Attributes:
        name (str)
        created_at (datetime)
        updated_at (datetime)
        deleted (bool)

        project (Relationship): `ToOne` relationship to Project
        created_by (Relationship): `ToOne` relationship to User

    """
    name = Field.String("name")
    created_at = Field.DateTime("created_at")
    updated_at = Field.DateTime("updated_at")
    size = Field.Int("size")
    consensus_settings = Field.Json("consensus_settings_json")

    # Relationships
    created_by = Relationship.ToOne("User")

    def __init__(self,
                 client,
                 project_id,
                 *args,
                 failed_data_row_ids=[],
                 **kwargs):
        super().__init__(client, *args, **kwargs)
        self.project_id = project_id
        self._failed_data_row_ids = failed_data_row_ids

    def project(self) -> 'Project':  # type: ignore
        """ Returns Project which this Batch belongs to

        Raises:
            LabelboxError: if the project is not found
        """
        query_str = """query getProjectPyApi($projectId: ID!) {
            project(
                where: {id: $projectId}){
                    %s
                }}""" % query.results_query_part(Entity.Project)
        params = {"projectId": self.project_id}
        response = self.client.execute(query_str, params)

        if response is None:
            raise ResourceNotFoundError(Entity.Project, params)

        return Entity.Project(self.client, response["project"])

    def remove_queued_data_rows(self) -> None:
        """ Removes remaining queued data rows from the batch and labeling queue.

        Args:
            batch (Batch): Batch to remove queued data rows from
        """

        project_id_param = "projectId"
        batch_id_param = "batchId"
        self.client.execute(
            """mutation RemoveQueuedDataRowsFromBatchPyApi($%s: ID!, $%s: ID!) {
            project(where: {id: $%s}) { removeQueuedDataRowsFromBatch(batchId: $%s) { id } }
        }""" % (project_id_param, batch_id_param, project_id_param,
                batch_id_param), {
                    project_id_param: self.project_id,
                    batch_id_param: self.uid
                },
            experimental=True)

    def export_data_rows(self,
                         timeout_seconds=120,
                         include_metadata: bool = False) -> Generator:
        """ Returns a generator that produces all data rows that are currently
        in this batch.

        Note: For efficiency, the data are cached for 30 minutes. Newly created data rows will not appear
        until the end of the cache period.

        Args:
            timeout_seconds (float): Max waiting time, in seconds.
            include_metadata (bool): True to return related DataRow metadata
        Returns:
            Generator that yields DataRow objects belonging to this batch.
        Raises:
            LabelboxError: if the export fails or is unable to download within the specified time.
        """
        warnings.warn(
            "You are currently utilizing exports v1 for this action, which will be deprecated after April 30th, 2024. We recommend transitioning to exports v2. To view export v2 details, visit our docs: https://docs.labelbox.com/reference/label-export",
            DeprecationWarning)

        id_param = "batchId"
        metadata_param = "includeMetadataInput"
        query_str = """mutation GetBatchDataRowsExportUrlPyApi($%s: ID!, $%s: Boolean!)
            {exportBatchDataRows(data:{batchId: $%s , includeMetadataInput: $%s}) {downloadUrl createdAt status}}
        """ % (id_param, metadata_param, id_param, metadata_param)
        sleep_time = 2
        while True:
            res = self.client.execute(query_str, {
                id_param: self.uid,
                metadata_param: include_metadata
            })
            res = res["exportBatchDataRows"]
            if res["status"] == "COMPLETE":
                download_url = res["downloadUrl"]
                response = requests.get(download_url)
                response.raise_for_status()
                reader = parser.reader(StringIO(response.text))
                return (
                    Entity.DataRow(self.client, result) for result in reader)
            elif res["status"] == "FAILED":
                raise LabelboxError("Data row export failed.")

            timeout_seconds -= sleep_time
            if timeout_seconds <= 0:
                raise LabelboxError(
                    f"Unable to export data rows within {timeout_seconds} seconds."
                )

            logger.debug("Batch '%s' data row export, waiting for server...",
                         self.uid)
            time.sleep(sleep_time)

    def delete(self) -> None:
        """ Deletes the given batch.

        Note: Batch deletion for batches that has labels is forbidden. 

        Args:
            batch (Batch): Batch to remove queued data rows from
        """

        project_id_param = "projectId"
        batch_id_param = "batchId"
        self.client.execute("""mutation DeleteBatchPyApi($%s: ID!, $%s: ID!) {
            project(where: {id: $%s}) { deleteBatch(batchId: $%s) { deletedBatchId } }
        }""" % (project_id_param, batch_id_param, project_id_param,
                batch_id_param), {
                    project_id_param: self.project_id,
                    batch_id_param: self.uid
                },
                            experimental=True)

    def delete_labels(self, set_labels_as_template=False) -> None:
        """ Deletes labels that were created for data rows in the batch.

        Args:
            batch (Batch): Batch to remove queued data rows from
            set_labels_as_template (bool): When set to true, the deleted labels will be kept as templates.
        """

        project_id_param = "projectId"
        batch_id_param = "batchId"
        type_param = "type"
        res = self.client.execute(
            """mutation DeleteBatchLabelsPyApi($%s: ID!, $%s: ID!, $%s: DeleteBatchLabelsType!) {
            project(where: {id: $%s}) { deleteBatchLabels(batchId: $%s, data:{ type: $%s }) { deletedLabelIds } }
        }""" % (project_id_param, batch_id_param, type_param, project_id_param,
                batch_id_param, type_param), {
                    project_id_param:
                        self.project_id,
                    batch_id_param:
                        self.uid,
                    type_param:
                        "RequeueDataWithLabelAsTemplate"
                        if set_labels_as_template else "RequeueData"
                },
            experimental=True)
        return res

    # modify this function to return an empty list if there are no failed data rows

    @property
    def failed_data_row_ids(self):
        if self._failed_data_row_ids is None:
            self._failed_data_row_ids = []

        return (x for x in self._failed_data_row_ids)



import datetime
from labelbox.schema.label import Label
import pytest
import uuid

from labelbox.data.annotation_types.data import (
    AudioData,
    ConversationData,
    DicomData,
    DocumentData,
    HTMLData,
    ImageData,
    TextData,
)
from labelbox.data.serialization import NDJsonConverter
from labelbox.data.annotation_types.data.video import VideoData

import labelbox as lb
import labelbox.types as lb_types
from labelbox.schema.media_type import MediaType
from labelbox.schema.annotation_import import AnnotationImportState
from labelbox import Project, Client

# Unit test for label based on data type.
# TODO: Dicom removed it is unstable when you deserialize and serialize on label import. If we intend to keep this library this needs add generic data types tests work with this data type.
# TODO: add MediaType.LLMPromptResponseCreation(data gen) once supported and llm human preference once media type is added


@pytest.mark.parametrize(
    "media_type, data_type_class",
    [
        (MediaType.Audio, AudioData),
        (MediaType.Html, HTMLData),
        (MediaType.Image, ImageData),
        (MediaType.Text, TextData),
        (MediaType.Video, VideoData),
        (MediaType.Conversational, ConversationData),
        (MediaType.Document, DocumentData),
    ],
)
def test_data_row_type_by_data_row_id(
    media_type,
    data_type_class,
    annotations_by_media_type,
    hardcoded_datarow_id,
):
    annotations_ndjson =  annotations_by_media_type[media_type]
    annotations_ndjson = [annotation[0] for annotation in annotations_ndjson]
    
    label = list(NDJsonConverter.deserialize(annotations_ndjson))[0]
   
    data_label = lb_types.Label(data=data_type_class(uid = hardcoded_datarow_id()),
                       annotations=label.annotations)

    assert data_label.data.uid == label.data.uid
    assert label.annotations == data_label.annotations


@pytest.mark.parametrize(
    "media_type, data_type_class",
    [
        (MediaType.Audio, AudioData),
        (MediaType.Html, HTMLData),
        (MediaType.Image, ImageData),
        (MediaType.Text, TextData),
        (MediaType.Video, VideoData),
        (MediaType.Conversational, ConversationData),
        (MediaType.Document, DocumentData),
    ],
)
def test_data_row_type_by_global_key(
    media_type,
    data_type_class,
    annotations_by_media_type,
    hardcoded_global_key,
):
    annotations_ndjson =  annotations_by_media_type[media_type]
    annotations_ndjson = [annotation[0] for annotation in annotations_ndjson]
    
    label = list(NDJsonConverter.deserialize(annotations_ndjson))[0]
   
    data_label = lb_types.Label(data=data_type_class(global_key = hardcoded_global_key()),
                       annotations=label.annotations)

    assert data_label.data.global_key == label.data.global_key
    assert label.annotations == data_label.annotations